# -*- coding: utf-8 -*-


import numpy as np
from numpy.testing import assert_array_almost_equal
import pytest

from africanus.model.wsclean.file_model import load
from africanus.model.wsclean.spec_model import (ordinary_spectral_model,
                                                log_spectral_model, spectra)
from africanus.model.wsclean.dask import spectra as dask_spectra


@pytest.fixture
def freq():
    return np.linspace(.856e9, .856e9*2, 16)


@pytest.fixture
def spectral_model_inputs(wsclean_model_file):
    sources = dict(load(wsclean_model_file))

    I, spi, log_si, ref_freq = (sources[n] for n in ("I", "SpectralIndex",
                                                     "LogarithmicSI",
                                                     "ReferenceFrequency"))

    I = np.asarray(I)  # noqa
    spi = np.asarray(spi)
    log_si = np.asarray(log_si)
    ref_freq = np.asarray(ref_freq)

    return I, spi, log_si, ref_freq


def test_spectral_model(spectral_model_inputs, freq):
    I, spi, log_si, ref_freq = spectral_model_inputs

    # Compute spectral model with numpy implementations
    ordinary_spec_model = ordinary_spectral_model(I, spi, log_si,
                                                  freq, ref_freq)
    log_spec_model = log_spectral_model(I, spi, log_si,
                                        freq, ref_freq)

    # Choose between ordinary and log spectral index
    # based on log_si array
    spec_model = np.where(log_si[:, None] == True,  # noqa
                          log_spec_model,
                          ordinary_spec_model)

    # Compare with our implementation
    model = spectra(I, spi, log_si, ref_freq, freq)
    assert_array_almost_equal(model, spec_model)

    # Ensure positive flux
    posI = I.copy()
    posI[I < 0.0] = -posI[I < 0.0]
    log_si_all_true = np.full(posI.shape, True, dtype=np.bool)
    log_spec_model = log_spectral_model(posI, spi, log_si_all_true,
                                        freq, ref_freq)
    model = spectra(posI, spi, True, ref_freq, freq)
    assert_array_almost_equal(model, log_spec_model)

    model = spectra(I, spi, False, ref_freq, freq)
    assert_array_almost_equal(model, ordinary_spec_model)


def test_dask_spectral_model(spectral_model_inputs, freq):
    da = pytest.importorskip("dask.array")

    I, spi, log_si, ref_freq = spectral_model_inputs

    # Compute spectral model with numpy implementations
    ordinary_spec_model = ordinary_spectral_model(I, spi, log_si,
                                                  freq, ref_freq)
    log_spec_model = log_spectral_model(I, spi, log_si,
                                        freq, ref_freq)

    # Choose between ordinary and log spectral index
    # based on log_si array
    spec_model = np.where(log_si[:, None] == True,  # noqa
                          log_spec_model,
                          ordinary_spec_model)

    # Create dask arrays
    src_chunks = (4, 3)
    spi_chunks = (2,)
    freq_chunks = (4, 4, 4, 4)

    I = da.from_array(I, chunks=(src_chunks,))  # noqa
    spi = da.from_array(spi, chunks=(src_chunks, spi_chunks))
    log_si = da.from_array(log_si, chunks=(src_chunks,))
    ref_freq = da.from_array(ref_freq, chunks=(src_chunks,))
    freq = da.from_array(freq, chunks=(freq_chunks,))

    # Compute spectra and compare
    model = dask_spectra(I, spi, log_si, ref_freq, freq)
    assert_array_almost_equal(model, spec_model)



# -*- coding: utf-8 -*-


from africanus.util.requirements import requires_optional

try:
    from astropy.coordinates import (EarthLocation, SkyCoord,
                                     AltAz, CIRS)
    from astropy.time import Time
    from astropy import units
except ImportError as e:
    astropy_import_error = e
    have_astropy_parangles = False
else:
    astropy_import_error = None
    have_astropy_parangles = True


@requires_optional('astropy', astropy_import_error)
def astropy_parallactic_angles(times, antenna_positions, field_centre):
    """
    Computes parallactic angles per timestep for the given
    reference antenna position and field centre.
    """
    ap = antenna_positions
    fc = field_centre

    # Convert from MJD second to MJD
    times = Time(times / 86400.00, format='mjd', scale='utc')

    ap = EarthLocation.from_geocentric(
        ap[:, 0], ap[:, 1], ap[:, 2], unit='m')
    fc = SkyCoord(ra=fc[0], dec=fc[1], unit=units.rad, frame='fk5')
    pole = SkyCoord(ra=0, dec=90, unit=units.deg, frame='fk5')

    cirs_frame = CIRS(obstime=times)
    pole_cirs = pole.transform_to(cirs_frame)
    fc_cirs = fc.transform_to(cirs_frame)

    altaz_frame = AltAz(location=ap[None, :], obstime=times[:, None])
    pole_altaz = pole_cirs[:, None].transform_to(altaz_frame)
    fc_altaz = fc_cirs[:, None].transform_to(altaz_frame)
    return fc_altaz.position_angle(pole_altaz)



# flake8: noqa

from africanus.rime.phase import phase_delay
from africanus.rime.feeds import feed_rotation
from africanus.rime.fast_beam_cubes import beam_cube_dde
from africanus.rime.parangles import parallactic_angles
from africanus.rime.transform import transform_sources
from africanus.rime.zernike import zernike_dde
from africanus.rime.predict import predict_vis, apply_gains



# -*- coding: utf-8 -*-


import logging

import numba
import numpy as np

try:
    import scipy.signal
    from scipy import optimize as opt
except ImportError as e:
    opt_import_err = e
else:
    opt_import_err = None

from africanus.util.requirements import requires_optional


@numba.jit(nopython=True, nogil=True, cache=True)
def twod_gaussian(coords, amplitude, xo, yo, sigma_x, sigma_y, theta, offset):
    x = coords[0]
    y = coords[1]
    xo = float(xo)
    yo = float(yo)
    a = (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_y**2)
    b = -(np.sin(2*theta))/(4*sigma_x**2) + (np.sin(2*theta))/(4*sigma_y**2)
    c = (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_y**2)
    g = (offset + amplitude *
         np.exp(- (a*((x-xo)**2) +
                   2*b*(x-xo)*(y-yo) +
                   c*((y-yo)**2))))
    return g.flatten()


@requires_optional('scipy', opt_import_err)
def fit_2d_gaussian(psf):
    """
    Fit an elliptical Gaussian to the primary lobe of the psf
    """

    # Get the full width at half maximum height of the psf

    # numba doesn't have argwhere, but it can jit argwhere's
    # implementation
    # I = np.stack((psf>=0.5*psf.max()).nonzero()).transpose()

    loc = np.argwhere(psf >= 0.5*psf.max())
    # Create an array with these values at the same indices and zeros otherwise
    lk, mk = psf.shape
    psf_fit = np.zeros_like(psf)
    psf_fit[loc[:, 0], loc[:, 1]] = psf[loc[:, 0], loc[:, 1]]
    # Create x and y indices
    x = np.linspace(0, psf.shape[0]-1, psf.shape[0])
    y = np.linspace(0, psf.shape[1]-1, psf.shape[1])
    x, y = np.meshgrid(x, y)
    # Set starting point of optimiser
    initial_guess = (0.5, lk/2, mk/2, 1.75, 1.4, -4.0, 0)
    # Flatten the data
    data = psf_fit.ravel()
    # Fit the function (Gaussian for now)
    popt, pcov = opt.curve_fit(twod_gaussian, (x, y), data, p0=initial_guess)
    # Get function with fitted params
    data_fitted = twod_gaussian((x, y), *popt)
    # Normalise the psf to have a max value of one
    data_fitted = data_fitted/data_fitted.max()
    return data_fitted.reshape(lk, mk)


@numba.jit(nopython=True, nogil=True, cache=True)
def find_peak(residuals):
    abs_residuals = residuals
    min_peak = abs_residuals.min()
    max_peak = abs_residuals.max()

    nx, ny = abs_residuals.shape

    minx, miny = -1, -1
    maxx, maxy = -1, -1
    peak_intensity = -1

    for x in range(nx):
        for y in range(ny):
            intensity = abs_residuals[x, y]

            if intensity == min_peak:
                minx = x
                miny = y

            if intensity == max_peak:
                maxx = x
                maxy = y
                peak_intensity = intensity

    if minx == -1 or miny == -1:
        raise ValueError("Minimum peak not found")

    if maxx == -1 or maxy == -1:
        raise ValueError("Maximum peak not found")

    return maxx, maxy, minx, miny, peak_intensity


@numba.jit(nopython=True, nogil=True, cache=True)
def build_cleanmap(clean, intensity, gamma, p, q):
    clean[p, q] += intensity*gamma


@numba.jit(nopython=True, nogil=True, cache=True)
def update_residual(residual, intensity, gamma, p, q, npix, psf):
    npix = residual.shape[0]  # Assuming square image
    residual -= gamma*intensity*psf[npix - 1 - p:2*npix - 1 - p,
                                    npix - 1 - q:2*npix - 1 - q]


def hogbom_clean(dirty, psf,
                 gamma=0.1,
                 threshold="default",
                 niter="default"):
    """
    Performs Hogbom Clean on the  ``dirty`` image given the ``psf``.

    Parameters
    ----------
    dirty : np.ndarray
        float64 dirty image of shape (ny, nx)
    psf : np.ndarray
        float64 Point Spread Function of shape (2*ny, 2*nx)
    gamma (optional) float
        the gain factor (must be less than one)
    threshold (optional) : float or str
        the threshold to clean to
    niter (optional : integer
        the maximum number of iterations allowed

    Returns
    -------
    np.ndarray
        float64 clean image of shape (ny, nx)
    np.ndarray
        float64 residual image of shape (ny, nx)
    """
    # deep copy dirties to first residuals,
    # want to keep the original dirty maps
    residuals = dirty.copy()

    # Check that psf is twice the size of residuals
    if (psf.shape[0] != 2*residuals.shape[0] or
            psf.shape[1] != 2*residuals.shape[1]):
        raise ValueError("Warning psf not right size")

    # Initialise array to store cleaned image
    clean = np.zeros_like(residuals)

    assert clean.shape[0] == clean.shape[1]
    npix = clean.shape[0]

    if niter == "default":
        niter = 3*npix

    p, q, pmin, qmin, intensity = find_peak(residuals)

    if threshold == "default":
        # Imin + 0.001*(intensity - Imin)
        threshold = 0.2*np.abs(intensity)
        logging.info("Threshold set at %s", threshold)
    else:
        # Imin + 0.001*(intensity - Imin)
        threshold = threshold*np.abs(intensity)
        logging.info("Assuming user set threshold at %s", threshold)

    # CLEAN the image
    i = 0

    while np.abs(intensity) > threshold and i <= niter:
        logging.info("min %f max %f peak %f threshold %f" %
                     (residuals.min(), residuals.max(), intensity, threshold))

        # First we set the
        build_cleanmap(clean, intensity, gamma, p, q)
        # Subtract out pixel
        update_residual(residuals, intensity, gamma, p, q, npix, psf)
        # Get new indices where residuals is max
        p, q, _, _, intensity = find_peak(residuals)
        # Increment counter
        i += 1
        # Warn if niter exceeded
        if i > niter:
            logging.warn("Number of iterations exceeded")
            logging.warn("Minimum residuals = %s", residuals.max())

    logging.info("Done cleaning after %d iterations.", i)

    return clean, residuals


@requires_optional("scipy", opt_import_err)
def restore(clean, psf, residuals):
    """
    Parameters
    ----------
    clean : np.ndarray
        float64 clean image of shape (ny, nx)
    psf : np.ndarray
        float64 Point Spread Function of shape (2*ny, 2*nx)
    residuals : np.ndarray
        float64 residual image of shape (ny, nx)

    Returns
    -------
    np.ndarray
        float64 Restored image of shape (ny, nx)
    np.ndarray
        float64 Convolved model of shape (ny, nx)
    """

    logging.info("Fitting 2D Gaussian")

    # get the ideal beam (fit 2D Gaussian to HWFH of psf)
    clean_beam = fit_2d_gaussian(psf)

    logging.info("Convolving")

    # cval=0.0) #Fast using fft
    iconv_model = scipy.signal.fftconvolve(clean, clean_beam, mode='same')

    logging.info("Convolving done")

    # Finally we add the residuals back to the image
    restored = iconv_model + residuals

    return (restored, iconv_model)


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

from functools import reduce
import logging
from operator import mul
from pathlib import Path

import numpy as np

from africanus.rime.fast_beam_cubes import BEAM_CUBE_DOCS
from africanus.util.code import format_code, memoize_on_key
from africanus.util.cuda import cuda_function, grids
from africanus.util.jinja2 import jinja_env
from africanus.util.requirements import requires_optional

try:
    import cupy as cp
    from cupy.core._scalar import get_typename as _get_typename
    from cupy.cuda.compiler import CompileException
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None


log = logging.getLogger(__name__)


_MAIN_TEMPLATE_PATH = Path("rime", "cuda", "beam.cu.j2")
_INTERP_TEMPLATE_PATH = Path("rime", "cuda", "beam_freq_interp.cu.j2")

BEAM_NUD_LIMIT = 128


def _freq_interp_key(beam_freq_map, frequencies):
    return (beam_freq_map.dtype, frequencies.dtype)


@memoize_on_key(_freq_interp_key)
def _generate_interp_kernel(beam_freq_map, frequencies):
    render = jinja_env.get_template(str(_INTERP_TEMPLATE_PATH)).render
    name = "beam_cube_freq_interp"

    block = (1024, 1, 1)

    code = render(kernel_name=name,
                  beam_nud_limit=BEAM_NUD_LIMIT,
                  blockdimx=block[0],
                  beam_freq_type=_get_typename(beam_freq_map.dtype),
                  freq_type=_get_typename(frequencies.dtype))

    code = code.encode('utf-8')
    dtype = np.result_type(beam_freq_map, frequencies)

    return cp.RawKernel(code, name), block, dtype


def _main_key_fn(beam, beam_lm_ext, beam_freq_map,
                 lm, parangles, pointing_errors,
                 antenna_scaling, frequencies,
                 dde_dims, ncorr):
    return (beam.dtype, beam.ndim, beam_lm_ext.dtype, beam_freq_map.dtype,
            lm.dtype, parangles.dtype, pointing_errors.dtype,
            antenna_scaling.dtype, frequencies.dtype, dde_dims, ncorr)


# Value to use in a bit shift to recover channel from flattened
# channel/correlation index
_corr_shifter = {4: 2, 2: 1, 1: 0}


@memoize_on_key(_main_key_fn)
def _generate_main_kernel(beam, beam_lm_ext, beam_freq_map,
                          lm, parangles, pointing_errors,
                          antenna_scaling, frequencies,
                          dde_dims, ncorr):

    beam_lw, beam_mh, beam_nud = beam.shape[:3]

    if beam_lw < 2 or beam_mh < 2 or beam_nud < 2:
        raise ValueError("(beam_lw, beam_mh, beam_nud) < 2 "
                         "to linearly interpolate")

    # Create template
    render = jinja_env.get_template(str(_MAIN_TEMPLATE_PATH)).render
    name = "beam_cube_dde"
    dtype = beam.dtype

    if dtype == np.complex64:
        block = (32, 32, 1)
    elif dtype == np.complex128:
        block = (32, 16, 1)
    else:
        raise TypeError("Need complex beam cube '%s'" % beam.dtype)

    try:
        corr_shift = _corr_shifter[ncorr]
    except KeyError:
        raise ValueError("Number of Correlations not in %s"
                         % list(_corr_shifter.keys()))

    coord_type = np.result_type(beam_lm_ext, lm, parangles,
                                pointing_errors, antenna_scaling,
                                np.float32)

    assert coord_type in (np.float32, np.float64)

    code = render(kernel_name=name,
                  blockdimx=block[0],
                  blockdimy=block[1],
                  blockdimz=block[2],
                  corr_shift=corr_shift,
                  ncorr=ncorr,
                  beam_nud_limit=BEAM_NUD_LIMIT,
                  # Beam type and manipulation functions
                  beam_type=_get_typename(beam.real.dtype),
                  beam_dims=beam.ndim,
                  make2_beam_fn=cuda_function('make2', beam.real.dtype),
                  beam_sqrt_fn=cuda_function('sqrt', beam.real.dtype),
                  beam_rsqrt_fn=cuda_function('rsqrt', beam.real.dtype),
                  # Coordinate type and manipulation functions
                  FT=_get_typename(coord_type),
                  floor_fn=cuda_function('floor', coord_type),
                  min_fn=cuda_function('min', coord_type),
                  max_fn=cuda_function('max', coord_type),
                  cos_fn=cuda_function('cos', coord_type),
                  sin_fn=cuda_function('sin', coord_type),
                  lm_ext_type=_get_typename(beam_lm_ext.dtype),
                  beam_freq_type=_get_typename(beam_freq_map.dtype),
                  lm_type=_get_typename(lm.dtype),
                  pa_type=_get_typename(parangles.dtype),
                  pe_type=_get_typename(pointing_errors.dtype),
                  as_type=_get_typename(antenna_scaling.dtype),
                  freq_type=_get_typename(frequencies.dtype),
                  dde_type=_get_typename(beam.real.dtype),
                  dde_dims=dde_dims)

    code = code.encode('utf-8')

    # Complex output type
    return cp.RawKernel(code, name), block, dtype


@requires_optional('cupy', opt_import_error)
def beam_cube_dde(beam, beam_lm_ext, beam_freq_map,
                  lm, parangles, pointing_errors,
                  antenna_scaling, frequencies):

    corrs = beam.shape[3:]

    if beam.shape[2] >= BEAM_NUD_LIMIT:
        raise ValueError("beam_nud exceeds %d" % BEAM_NUD_LIMIT)

    nsrc = lm.shape[0]
    ntime, na = parangles.shape
    nchan = frequencies.shape[0]
    ncorr = reduce(mul, corrs, 1)
    nchancorr = nchan*ncorr

    oshape = (nsrc, ntime, na, nchan) + corrs

    if len(corrs) > 1:
        # Flatten the beam correlation dims
        fbeam = beam.reshape(beam.shape[:3] + (ncorr,))
    else:
        fbeam = beam

    # Generate frequency interpolation kernel
    ikernel, iblock, idt = _generate_interp_kernel(beam_freq_map, frequencies)

    # Generate main beam cube kernel
    kernel, block, dtype = _generate_main_kernel(fbeam, beam_lm_ext,
                                                 beam_freq_map,
                                                 lm, parangles,
                                                 pointing_errors,
                                                 antenna_scaling,
                                                 frequencies,
                                                 len(oshape),
                                                 ncorr)
    # Call frequency interpolation kernel
    igrid = grids((nchan, 1, 1), iblock)
    freq_data = cp.empty((3, nchan), dtype=frequencies.dtype)

    try:
        ikernel(igrid, iblock, (frequencies, beam_freq_map, freq_data))
    except CompileException:
        log.exception(format_code(ikernel.code))
        raise

    # Call main beam cube kernel
    out = cp.empty((nsrc, ntime, na, nchan) + (ncorr,), dtype=beam.dtype)
    grid = grids((nchancorr, na, ntime), block)

    try:
        kernel(grid, block, (fbeam, beam_lm_ext, beam_freq_map,
                             lm, parangles, pointing_errors,
                             antenna_scaling, frequencies, freq_data,
                             nsrc, out))
    except CompileException:
        log.exception(format_code(kernel.code))
        raise

    return out.reshape(oshape)


try:
    beam_cube_dde.__doc__ = BEAM_CUBE_DOCS.substitute(
                                array_type=":class:`cupy.ndarray`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

import numpy as np

try:
    import dask
    import dask.array as da
    from dask.base import normalize_token
    from dask.highlevelgraph import HighLevelGraph
except ImportError as e:
    import_error = e
else:
    import_error = None

try:
    import nifty_gridder as ng
except ImportError:
    nifty_import_err = ImportError("Please manually install nifty_gridder "
                                   "from https://gitlab.mpcdf.mpg.de/ift/"
                                   "nifty_gridder.git")
else:
    nifty_import_err = None

from africanus.util.requirements import requires_optional


class GridderConfigWrapper(object):
    """
    Wraps a nifty GridderConfiguration for pickling purposes.
    """

    def __init__(self, nx=1024, ny=1024, eps=2e-13,
                 cell_size_x=2.0, cell_size_y=2.0):
        self.nx = nx
        self.ny = ny
        self.csx = cell_size_x
        self.csy = cell_size_y
        self.eps = eps
        self.grid_config = ng.GridderConfig(nx, ny, eps,
                                            cell_size_x,
                                            cell_size_y)

    @property
    def object(self):
        return self.grid_config

    def __reduce__(self):
        return (GridderConfigWrapper,
                (self.nx, self.ny, self.eps, self.csx, self.csy))


if import_error is None:
    @normalize_token.register(GridderConfigWrapper)
    def normalize_gridder_config_wrapper(gc):
        return normalize_token((gc.nx, gc.ny, gc.csx, gc.csy, gc.eps))


@requires_optional("dask.array", import_error)
@requires_optional("nifty_gridder", nifty_import_err)
def grid_config(nx=1024, ny=1024, eps=2e-13, cell_size_x=2.0, cell_size_y=2.0):
    """
    Returns a wrapper around a NIFTY GridderConfiguration object.


    Parameters
    ----------
    nx : int, optional
        Number of X pixels in the grid. Defaults to 1024.
    ny : int, optional
        Number of Y pixels in the grid. Defaults to 1024.
    cell_size_x : float, optional
        Cell size of the X pixel in arcseconds. Defaults to 2.0.
    cell_size_y : float, optional
        Cell size of the Y pixel in arcseconds. Defaults to 2.0.
    eps : float
        Gridder accuracy error. Defaults to 2e-13


    Returns
    -------
    grid_config : :class:`GridderConfigWrapper`
        The NIFTY Gridder Configuration
    """
    return GridderConfigWrapper(nx, ny, eps, cell_size_x, cell_size_y)


def _nifty_baselines(uvw, chan_freq):
    """ Wrapper function for creating baseline mappings per row chunk """
    assert len(chan_freq) == 1, "Handle multiple channel chunks"
    return ng.Baselines(uvw[0], chan_freq[0])


def _nifty_indices(baselines, grid_config, flag,
                   chan_begin, chan_end, wmin, wmax):
    """ Wrapper function for creating indices per row chunk """
    return ng.getIndices(baselines, grid_config, flag[0],
                         chan_begin, chan_end, wmin, wmax)


def _nifty_grid(baselines, grid_config, indices, vis, weights):
    """ Wrapper function for creating a grid of visibilities per row chunk """
    assert len(vis) == 1 and type(vis) == list
    return ng.ms2grid_c(baselines, grid_config, indices,
                        vis[0], None, weights[0])[None, :, :]


def _nifty_grid_streams(baselines, grid_config, indices,
                        vis, weights, grid_in=None):
    """ Wrapper function for creating a grid of visibilities per row chunk """
    return ng.ms2grid_c(baselines, grid_config, indices,
                        vis, grid_in, weights)


class GridStreamReduction(Mapping):
    """
    tl;dr this is a dictionary that is expanded in place when
    first accessed. Saves memory when pickled for sending
    to the dask scheduler.

    See :class:`dask.blockwise.Blockwise` for further insight.

    Produces graph serially summing coherencies in
    ``stream`` parallel streams.
    """

    def __init__(self, baselines, indices, gc,
                 corr_vis, corr_weights,
                 corr, streams):
        token = dask.base.tokenize(baselines, indices, gc,
                                   corr_vis, corr_weights,
                                   corr, streams)
        self.name = "-".join(("nifty-grid-stream", str(corr), token))
        self.bl_name = baselines.name
        self.idx_name = indices.name
        self.cvis_name = corr_vis.name
        self.wgt_name = corr_weights.name
        self.gc = gc
        self.corr = corr

        self.row_blocks = indices.numblocks[0]
        self.streams = streams

    @property
    def _dict(self):
        if hasattr(self, "_cached_dict"):
            return self._cached_dict
        else:
            self._cached_dict = self._create_dict()
            return self._cached_dict

    def __getitem__(self, key):
        return self._dict[key]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        # Extract dimension blocks
        return self.row_blocks

    def _create_dict(self):
        # Graph dictionary
        layers = {}

        name = self.name
        gc = self.gc
        row_blocks = self.row_blocks
        streams = self.streams
        indices_name = self.idx_name
        baselines_name = self.bl_name
        corr_vis_name = self.cvis_name
        corr_wgt_name = self.wgt_name

        cb = 0  # Assume single channel

        # Split our row blocks by the number of streams
        # For all blocks in a stream, we'll grid those
        # blocks serially, passing one grid into the other
        row_block_chunks = (row_blocks + streams - 1) // streams

        for rb_start in range(0, row_blocks, row_block_chunks):
            rb_end = min(rb_start + row_block_chunks, row_blocks)
            last_key = None

            for rb in range(rb_start, rb_end):
                fn = (_nifty_grid_streams,
                      (baselines_name, rb),
                      gc,
                      (indices_name, rb),
                      (corr_vis_name, rb, cb),
                      (corr_wgt_name, rb, cb),
                      # Re-use grid from last operation if present
                      last_key)

                key = (name, rb, cb)
                layers[key] = fn
                last_key = key

        return layers


class FinalGridReduction(Mapping):
    """
    tl;dr this is a dictionary that is expanded in place when
    first accessed. Saves memory when pickled for sending
    to the dask scheduler.

    See :class:`dask.blockwise.Blockwise` for further insight.

    Produces graph serially summing coherencies in
    ``stream`` parallel streams.
    """

    def __init__(self, grid_stream_reduction):
        self.in_name = grid_stream_reduction.name
        token = dask.base.tokenize(grid_stream_reduction)
        self.name = "grid-stream-reduction-" + token
        self.row_blocks = grid_stream_reduction.row_blocks
        self.streams = grid_stream_reduction.streams

    @property
    def _dict(self):
        if hasattr(self, "_cached_dict"):
            return self._cached_dict
        else:
            self._cached_dict = self._create_dict()
            return self._cached_dict

    def __getitem__(self, key):
        return self._dict[key]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        # Always returns a single block, corresponding to one image
        return 1

    def _create_dict(self):
        # Graph dictionary
        layers = {}

        row_blocks = self.row_blocks
        streams = self.streams
        cb = 0  # Assume single channel

        # Split our row blocks by the number of streams
        # For all blocks in a stream, we'll grid those
        # blocks serially, passing one grid into the other
        row_block_chunks = (row_blocks + streams - 1) // streams
        last_keys = []

        for ob, rb_start in enumerate(range(0, row_blocks, row_block_chunks)):
            rb_end = min(rb_start + row_block_chunks, row_blocks)
            last_keys.append((self.in_name, rb_end - 1, cb))

        key = (self.name, 0, 0)
        task = (sum, last_keys)
        layers[key] = task

        return layers


@requires_optional("dask.array", import_error)
@requires_optional("nifty_gridder", nifty_import_err)
def grid(vis, uvw, flags, weights, frequencies, grid_config,
         wmin=-1e30, wmax=1e30, streams=None):
    """
    Grids the supplied visibilities in parallel. Note that
    a grid is create for each visibility chunk.

    Parameters
    ----------
    vis : :class:`dask.array.Array`
        visibilities of shape :code:`(row, chan, corr)`
    uvw : :class:`dask.array.Array`
        uvw coordinates of shape :code:`(row, 3)`
    flags : :class:`dask.array.Array`
        flags of shape :code:`(row, chan, corr)`
    weights : :class:`dask.array.Array`
        weights of shape :code:`(row, chan, corr)`.
    frequencies : :class:`dask.array.Array`
        frequencies of shape :code:`(chan,)`
    grid_config : :class:`GridderConfigWrapper`
        Gridding Configuration
    wmin : float
        Minimum W coordinate to grid. Defaults to -1e30.
    wmax : float
        Maximum W coordinate to grid. Default to 1e30.
    streams : int, optional
        Number of parallel gridding operations. Default to None,
        in which case as many grids as visibility chunks will
        be created.

    Returns
    -------
    grid : :class:`dask.array.Array`
        grid of shape :code:`(ny, nx, corr)`
    """
    if len(frequencies.chunks[0]) != 1:
        raise ValueError("Chunking in channel currently unsupported")

    # Create a baseline object per row chunk
    baselines = da.blockwise(_nifty_baselines, ("row",),
                             uvw, ("row", "uvw"),
                             frequencies, ("chan",),
                             dtype=np.object)

    if len(frequencies.chunks[0]) != 1:
        raise ValueError("Chunking in channel unsupported")

    gc = grid_config.object
    grids = []

    for corr in range(vis.shape[2]):
        corr_flags = flags[:, :, corr]
        corr_vis = vis[:, :, corr]
        corr_weights = weights[:, :, corr]

        indices = da.blockwise(_nifty_indices, ("row",),
                               baselines, ("row",),
                               gc, None,
                               corr_flags, ("row", "chan"),
                               -1, None,  # channel begin
                               -1, None,  # channel end
                               wmin, None,
                               wmax, None,
                               dtype=np.int32)

        if streams is None:
            # Standard parallel reduction, possibly memory hungry
            # if many threads (and thus grids) are gridding
            # parallel
            grid = da.blockwise(_nifty_grid, ("row", "nu", "nv"),
                                baselines, ("row",),
                                gc, None,
                                indices, ("row",),
                                corr_vis, ("row", "chan"),
                                corr_weights, ("row", "chan"),
                                new_axes={"nu": gc.Nu(), "nv": gc.Nv()},
                                adjust_chunks={"row": 1},
                                dtype=np.complex128)

            grids.append(grid.sum(axis=0))
        else:
            # Stream reduction
            layers = GridStreamReduction(baselines, indices, gc,
                                         corr_vis, corr_weights,
                                         corr, streams)
            deps = [baselines, indices, corr_vis, corr_weights]
            graph = HighLevelGraph.from_collections(layers.name, layers, deps)
            chunks = corr_vis.chunks
            grid_stream_red = da.Array(graph, layers.name, chunks, vis.dtype)

            layers = FinalGridReduction(layers)
            deps = [grid_stream_red]
            graph = HighLevelGraph.from_collections(layers.name, layers, deps)
            chunks = ((gc.Nu(),), (gc.Nv(),))
            corr_grid = da.Array(graph, layers.name, chunks, vis.dtype)

            grids.append(corr_grid)

    return da.stack(grids, axis=2)


def _nifty_dirty(grid, grid_config):
    """ Wrapper function for creating a dirty image """
    grids = [grid_config.grid2dirty_c(grid[:, :, c]).real
             for c in range(grid.shape[2])]

    return np.stack(grids, axis=2)


@requires_optional("dask.array", import_error)
@requires_optional("nifty_gridder", nifty_import_err)
def dirty(grid, grid_config):
    """
    Computes the dirty image from gridded visibilities and the
    gridding configuration.

    Parameters
    ----------
    grid : :class:`dask.array.Array`
        Gridded visibilities of shape :code:`(nv, nu, ncorr)`
    grid_config : :class:`GridderConfigWrapper`
        Gridding configuration

    Returns
    -------
    dirty : :class:`dask.array.Array`
        dirty image of shape :code:`(ny, nx, corr)`
    """

    gc = grid_config.object
    nx = gc.Nxdirty()
    ny = gc.Nydirty()

    return da.blockwise(_nifty_dirty, ("nx", "ny", "corr"),
                        grid, ("nx", "ny", "corr"),
                        gc, None,
                        adjust_chunks={"nx": nx, "ny": ny},
                        dtype=grid.real.dtype)


def _nifty_model(image, grid_config):
    """ Wrapper function for creating a dirty image """
    images = [grid_config.dirty2grid_c(image[:, :, c])
              for c in range(image.shape[2])]

    return np.stack(images, axis=2)


@requires_optional("dask.array", import_error)
@requires_optional("nifty_gridder", nifty_import_err)
def model(image, grid_config):
    """
    Computes model visibilities from an image
    and a gridding configuration.

    Parameters
    ----------
    image : :class:`dask.array.Array`
        Image of shape :code:`(ny, nx, corr)`.
    grid_config : :class:`GridderConfigWrapper`
        nifty gridding configuration object

    Returns
    -------
    model_vis : :class:`dask.array.Array`
        Model visibilities of shape :code:`(nu, nv, corr)`.
    """

    gc = grid_config.object
    nu = gc.Nu()
    nv = gc.Nv()

    return da.blockwise(_nifty_model, ("nu", "nv", "corr"),
                        image, ("nu", "nv", "corr"),
                        gc, None,
                        adjust_chunks={"nu": nu, "nv": nv},
                        dtype=image.dtype)


def _nifty_degrid(grid, baselines, indices, grid_config):
    assert len(grid) == 1 and len(grid[0]) == 1
    return ng.grid2ms_c(baselines, grid_config.object, indices, grid[0][0])


@requires_optional("dask.array", import_error)
@requires_optional("nifty_gridder", nifty_import_err)
def degrid(grid, uvw, flags, weights, frequencies,
           grid_config, wmin=-1e30, wmax=1e30):
    """
    Degrids the visibilities from the supplied grid in parallel.

    Parameters
    ----------
    grid : :class:`dask.array.Array`
        gridded visibilities of shape :code:`(ny, nx, corr)`
    uvw : :class:`dask.array.Array`
        uvw coordinates of shape :code:`(row, 3)`
    flags : :class:`dask.array.Array`
        flags of shape :code:`(row, chan, corr)`
    weights : :class:`dask.array.Array`
        weights of shape :code:`(row, chan, corr)`.
        Currently unsupported and ignored.
    frequencies : :class:`dask.array.Array`
        frequencies of shape :code:`(chan,)`
    grid_config : :class:`GridderConfigWrapper`
        Gridding Configuration
    wmin : float
        Minimum W coordinate to grid. Defaults to -1e30.
    wmax : float
        Maximum W coordinate to grid. Default to 1e30.

    Returns
    -------
    grid : :class:`dask.array.Array`
        grid of shape :code:`(ny, nx, corr)`
    """

    if len(frequencies.chunks[0]) != 1:
        raise ValueError("Chunking in channel currently unsupported")

    # Create a baseline object per row chunk
    baselines = da.blockwise(_nifty_baselines, ("row",),
                             uvw, ("row", "uvw"),
                             frequencies, ("chan",),
                             dtype=np.object)

    gc = grid_config.object
    vis_chunks = []

    for corr in range(grid.shape[2]):
        corr_flags = flags[:, :, corr].map_blocks(np.require, requirements="C")
        corr_grid = grid[:, :, corr].map_blocks(np.require, requirements="C")

        indices = da.blockwise(_nifty_indices, ("row",),
                               baselines, ("row",),
                               gc, None,
                               corr_flags, ("row", "chan"),
                               -1, None,  # channel begin
                               -1, None,  # channel end
                               wmin, None,
                               wmax, None,
                               dtype=np.int32)

        vis = da.blockwise(_nifty_degrid, ("row", "chan"),
                           corr_grid, ("ny", "nx"),
                           baselines, ("row",),
                           indices, ("row",),
                           grid_config, None,
                           new_axes={"chan": frequencies.shape[0]},
                           dtype=grid.dtype)

        vis_chunks.append(vis)

    return da.stack(vis_chunks, axis=2)



# -*- coding: utf-8 -*-


from collections import OrderedDict, deque
from pprint import pformat
from textwrap import fill

import numpy as np

from africanus.util.casa_types import (STOKES_TYPES,
                                       STOKES_ID_MAP)
from africanus.util.docs import DocstringTemplate

stokes_conv = {
    'RR': {('I', 'V'): lambda i, v: i + v + 0j},
    'RL': {('Q', 'U'): lambda q, u: q + u*1j},
    'LR': {('Q', 'U'): lambda q, u: q - u*1j},
    'LL': {('I', 'V'): lambda i, v: i - v + 0j},

    'XX': {('I', 'Q'): lambda i, q: i + q + 0j},
    'XY': {('U', 'V'): lambda u, v: u + v*1j},
    'YX': {('U', 'V'): lambda u, v: u - v*1j},
    'YY': {('I', 'Q'): lambda i, q: i - q + 0j},

    'I': {('XX', 'YY'): lambda xx, yy: (xx + yy).real / 2,
          ('RR', 'LL'): lambda rr, ll: (rr + ll).real / 2},

    'Q': {('XX', 'YY'): lambda xx, yy: (xx - yy).real / 2,
          ('RL', 'LR'): lambda rl, lr: (rl + lr).real / 2},

    'U': {('XY', 'YX'): lambda xy, yx: (xy + yx).real / 2,
          ('RL', 'LR'): lambda rl, lr: (rl - lr).imag / 2},

    'V': {('XY', 'YX'): lambda xy, yx: (xy - yx).imag / 2,
          ('RR', 'LL'): lambda rr, ll: (rr - ll).real / 2},
}


class DimensionMismatch(Exception):
    pass


class MissingConversionInputs(Exception):
    pass


def _element_indices_and_shape(data):
    if not isinstance(data, (tuple, list)):
        data = [data]

    # Shape of the data
    shape = []

    # Each stack element is (list, index, depth)
    queue = deque([(data, (), 0)])
    result = OrderedDict()

    while len(queue) > 0:
        current, current_idx, depth = queue.popleft()

        # First do shape inference
        if len(shape) <= depth:
            shape.append(len(current))
        elif shape[depth] != len(current):
            raise DimensionMismatch("Dimension mismatch %d != %d at depth %d"
                                    % (shape[depth], len(current), depth))

        # Handle each sequence element
        for i, e in enumerate(current):
            # Found a list, recurse
            if isinstance(e, (tuple, list)):
                queue.append((e, current_idx + (i, ), depth + 1))
            # String
            elif isinstance(e, str):
                if e in result:
                    raise ValueError("'%s' defined multiple times" % e)

                result[e] = current_idx + (i, )
            # We have a CASA integer Stokes ID, convert to string
            elif np.issubdtype(type(e), np.integer):
                try:
                    e = STOKES_ID_MAP[e]
                except KeyError:
                    raise ValueError("Invalid id '%d'. "
                                     "Valid id's '%s'"
                                     % (e, pformat(STOKES_ID_MAP)))

                if e in result:
                    raise ValueError("'%s' defined multiple times" % e)

                result[e] = current_idx + (i, )
            else:
                raise TypeError("Invalid type '%s' for element '%s'"
                                % (type(e), e))

    return result, tuple(shape)


def convert_setup(input, input_schema, output_schema):
    input_indices, input_shape = _element_indices_and_shape(input_schema)
    output_indices, output_shape = _element_indices_and_shape(output_schema)

    if input.shape[-len(input_shape):] != input_shape:
        raise ValueError("Last dimension of input doesn't match input schema")

    mapping = []
    dummy = input.dtype.type(0)

    # Figure out how to produce an output from available inputs
    for okey, out_idx in output_indices.items():
        try:
            deps = stokes_conv[okey]
        except KeyError:
            raise ValueError("Unknown output '%s'. Known types '%s'"
                             % (deps, STOKES_TYPES))

        found_conv = False

        # Find a mapping for which we have inputs
        for (c1, c2), fn in deps.items():
            # Get indices for both correlations
            try:
                c1_idx = (Ellipsis,) + input_indices[c1]
            except KeyError:
                continue

            try:
                c2_idx = (Ellipsis,) + input_indices[c2]
            except KeyError:
                continue

            found_conv = True
            out_idx = (Ellipsis,) + out_idx
            # Figure out the data type for this output
            dtype = fn(dummy, dummy).dtype
            mapping.append((c1_idx, c2_idx, out_idx, fn, dtype))
            break

        # We must find a conversion
        if not found_conv:
            raise MissingConversionInputs("None of the supplied inputs '%s' "
                                          "can produce output '%s'. It can be "
                                          "produced by the following "
                                          "combinations '%s'." % (
                                                input_schema,
                                                okey, deps.keys()))

    out_dtype = np.result_type(*[dt for _, _, _, _, dt in mapping])

    return mapping, input_shape, output_shape, out_dtype


def convert_impl(input, mapping, in_shape, out_shape, dtype):
    # Make the output array
    out_shape = input.shape[:-len(in_shape)] + out_shape
    output = np.empty(out_shape, dtype=dtype)

    for c1_idx, c2_idx, out_idx, fn, _ in mapping:
        output[out_idx] = fn(input[c1_idx], input[c2_idx])

    return output


def convert(input, input_schema, output_schema):
    """ See STOKES_DOCS below """

    # Do the conversion
    mapping, in_shape, out_shape, dtype = convert_setup(input,
                                                        input_schema,
                                                        output_schema)

    return convert_impl(input, mapping, in_shape, out_shape, dtype)


CONVERT_DOCS = """
This function converts forward and backward
from stokes ``I,Q,U,V`` to both linear ``XX,XY,YX,YY``
and circular ``RR, RL, LR, LL`` correlations.

For example, we can convert from stokes parameters
to linear correlations:

.. code-block:: python

    stokes.shape == (10, 4, 4)
    corrs = convert(stokes, ["I", "Q", "U", "V"],
                    [['XX', 'XY'], ['YX', 'YY'])

    assert corrs.shape == (10, 4, 2, 2)

Or circular correlations to stokes:

.. code-block:: python

    vis.shape == (10, 4, 2, 2)

    stokes = convert(vis, [['RR', 'RL'], ['LR', 'LL']],
                            ['I', 'Q', 'U', 'V'])

    assert stokes.shape == (10, 4, 4)

``input`` can ``output`` can be arbitrarily nested or ordered lists,
but the appropriate inputs must be present to produce the requested
outputs.

The elements of ``input`` and ``output`` may be strings or integers
representing stokes parameters or correlations. See the Notes
for a full list.


Notes
-----

Only stokes parameters, linear and circular correlations are
currently handled, but the full list of id's and strings as defined
in the `CASA documentation
<https://casacore.github.io/casacore/classcasacore_1_1Stokes.html>`_
is:

.. code-block:: python

    {stokes_type_map}

Parameters
----------
input : $(array_type)
    Complex or floating point input data of shape
    :code:`(dim_1, ..., dim_n, icorr_1, ..., icorr_m)`
input_schema : list of str or int
    A schema describing the :code:`icorr_1, ..., icorr_m`
    dimension of ``input``. Must have the same shape as
    the last dimensions of ``input``.
output_schema : list of str or int
    A schema describing the :code:`ocorr_1, ..., ocorr_n`
    dimension of the return value.

Returns
-------
 result : $(array_type)
    Result of shape :code:`(dim_1, ..., dim_n, ocorr_1, ..., ocorr_m)`
    The type may be floating point or promoted to complex
    depending on the combinations in ``output``.
"""

# Fill in the STOKES TYPES
_map_str = ", ".join(["%s: %d" % (t, i) for i, t in enumerate(STOKES_TYPES)])
_map_str = "{{ " + _map_str + " }}"
# Indent must match docstrings
_map_str = fill(_map_str, initial_indent='', subsequent_indent=' '*8)
CONVERT_DOCS = DocstringTemplate(CONVERT_DOCS.format(stokes_type_map=_map_str))
del _map_str

try:
    convert.__doc__ = CONVERT_DOCS.substitute(
                                  array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


import ast
import contextlib
import logging
import os
import re
import shutil
import subprocess
import sys
import tempfile

from os.path import join as pjoin
from pkg_resources import resource_filename

import distutils
from distutils import errors
from distutils import msvccompiler
from distutils import unixccompiler

from africanus.util.code import format_code
from africanus.util.requirements import requires_optional

try:
    import cupy as cp
except ImportError as e:
    cupy_import_error = e
else:
    cupy_import_error = None


log = logging.getLogger(__name__)


def get_path(key):
    return os.environ.get(key, '').split(os.pathsep)


def search_on_path(filenames):
    for p in get_path('PATH'):
        for filename in filenames:
            full = os.path.join(p, filename)
            if os.path.exists(full):
                return os.path.abspath(full)
    return None


PLATFORM_DARWIN = sys.platform.startswith('darwin')
PLATFORM_LINUX = sys.platform.startswith('linux')
PLATFORM_WIN32 = sys.platform.startswith('win32')

minimum_cuda_version = 8000
minimum_cudnn_version = 5000
maximum_cudnn_version = 7999

_cuda_path = 'NOT_INITIALIZED'
_compiler_base_options = None
_cuda_info = None


@contextlib.contextmanager
def _tempdir():
    temp_dir = tempfile.mkdtemp()
    try:
        yield temp_dir
    finally:
        shutil.rmtree(temp_dir, ignore_errors=True)


def get_cuda_path():
    global _cuda_path

    # Use a magic word to represent the cache not filled because None is a
    # valid return value.
    if _cuda_path != 'NOT_INITIALIZED':
        return _cuda_path

    nvcc_path = search_on_path(('nvcc', 'nvcc.exe'))
    cuda_path_default = None
    if nvcc_path is None:
        log.warn('nvcc not in path. Please set path to nvcc.')
    else:
        cuda_path_default = os.path.normpath(
            os.path.join(os.path.dirname(nvcc_path), '..'))

    cuda_path = os.environ.get('CUDA_PATH', '')  # Nvidia default on Windows
    if len(cuda_path) > 0 and cuda_path != cuda_path_default:
        log.warn('nvcc path != CUDA_PATH')
        log.warn('nvcc path: %s' % cuda_path_default)
        log.warn('CUDA_PATH: %s' % cuda_path)

    if os.path.exists(cuda_path):
        _cuda_path = cuda_path
    elif cuda_path_default is not None:
        _cuda_path = cuda_path_default
    elif os.path.exists('/usr/local/cuda'):
        _cuda_path = '/usr/local/cuda'
    else:
        _cuda_path = None

    return _cuda_path


def get_nvcc_path():
    nvcc = os.environ.get('NVCC', None)
    if nvcc:
        return distutils.split_quoted(nvcc)

    cuda_path = get_cuda_path()
    if cuda_path is None:
        return None

    if PLATFORM_WIN32:
        nvcc_bin = 'bin/nvcc.exe'
    else:
        nvcc_bin = 'bin/nvcc'

    nvcc_path = os.path.join(cuda_path, nvcc_bin)
    if os.path.exists(nvcc_path):
        return [nvcc_path]
    else:
        return None


def get_compiler_setting():
    cuda_path = get_cuda_path()

    include_dirs = []
    library_dirs = []
    define_macros = []

    if cuda_path:
        include_dirs.append(os.path.join(cuda_path, 'include'))
        if PLATFORM_WIN32:
            library_dirs.append(os.path.join(cuda_path, 'bin'))
            library_dirs.append(os.path.join(cuda_path, 'lib', 'x64'))
        else:
            library_dirs.append(os.path.join(cuda_path, 'lib64'))
            library_dirs.append(os.path.join(cuda_path, 'lib'))
    if PLATFORM_DARWIN:
        library_dirs.append('/usr/local/cuda/lib')

    if PLATFORM_WIN32:
        nvtoolsext_path = os.environ.get('NVTOOLSEXT_PATH', '')
        if os.path.exists(nvtoolsext_path):
            include_dirs.append(os.path.join(nvtoolsext_path, 'include'))
            library_dirs.append(os.path.join(nvtoolsext_path, 'lib', 'x64'))
        else:
            define_macros.append(('CUPY_NO_NVTX', '1'))

    return {
        'include_dirs': include_dirs,
        'library_dirs': library_dirs,
        'define_macros': define_macros,
        'language': 'c++',
    }


def _match_output_lines(output_lines, regexs):
    # Matches regular expressions `regexs` against `output_lines` and finds the
    # consecutive matching lines from `output_lines`.
    # `None` is returned if no match is found.
    if len(output_lines) < len(regexs):
        return None

    matches = [None] * len(regexs)
    for i in range(len(output_lines) - len(regexs)):
        for j in range(len(regexs)):
            m = re.match(regexs[j], output_lines[i + j])
            if not m:
                break
            matches[j] = m
        else:
            # Match found
            return matches

    # No match
    return None


def get_compiler_base_options():
    """Returns base options for nvcc compiler.

    """
    global _compiler_base_options
    if _compiler_base_options is None:
        _compiler_base_options = _get_compiler_base_options()
    return _compiler_base_options


def _get_compiler_base_options():
    # Try compiling a dummy code.
    # If the compilation fails, try to parse the output of compilation
    # and try to compose base options according to it.
    nvcc_path = get_nvcc_path()
    with _tempdir() as temp_dir:
        test_cu_path = os.path.join(temp_dir, 'test.cu')
        test_out_path = os.path.join(temp_dir, 'test.out')
        with open(test_cu_path, 'w') as f:
            f.write('int main() { return 0; }')
        proc = subprocess.Popen(
            nvcc_path + ['-o', test_out_path, test_cu_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        stdoutdata, stderrdata = proc.communicate()
        stderrlines = stderrdata.split(b'\n')
        if proc.returncode != 0:

            # No supported host compiler
            matches = _match_output_lines(
                stderrlines,
                [
                    b'^ERROR: No supported gcc/g\\+\\+ host compiler found, '
                    b'but .* is available.$',
                    b'^ *Use \'nvcc (.*)\' to use that instead.$',
                ])
            if matches is not None:
                base_opts = matches[1].group(1)
                base_opts = base_opts.decode('utf8').split(' ')
                return base_opts

            # Unknown error
            raise RuntimeError(
                'Encountered unknown error while testing nvcc:\n' +
                stderrdata.decode('utf8'))

    return []


def _get_cuda_info():
    nvcc_path = get_nvcc_path()

    code = '''
    #include <cuda.h>
    #include <stdio.h>
    int main(int argc, char* argv[]) {
        int nDevices;
        cudaGetDeviceCount(&nDevices);

        printf("{\\n");
        printf("'cuda_version': %d,\\n", CUDA_VERSION);
        printf("'devices': [\\n");

        for(int d=0; d < nDevices; ++d) {
            cudaDeviceProp props;
            cudaGetDeviceProperties(&props, d);

            printf("{\\n");
            printf("'name'                  :'%s',\\n", props.name);
            printf("'major'                 :  %d,\\n", props.major);
            printf("'minor'                 :  %d,\\n", props.minor);
            printf("'total_global_mem'      :  %u,\\n", props.totalGlobalMem);
            printf("'warp_size'             :  %u,\\n", props.warpSize);
            printf("'max_threads_per_block' :  %u,\\n", props.maxThreadsPerBlock);
            printf("'max_thread_size'       : (%u,%u,%u),\\n",
                                                        props.maxThreadsDim[0],
                                                        props.maxThreadsDim[1],
                                                        props.maxThreadsDim[2]);
            printf("'max_grid_size'         : (%u,%u,%u),\\n",
                                                        props.maxGridSize[0],
                                                        props.maxGridSize[1],
                                                        props.maxGridSize[2]);
            printf("'device_overlap'        :  %s,\\n", props.deviceOverlap ? "True" : "False");
            printf("'async_engines'         :  %u,\\n", props.asyncEngineCount);
            printf("'multiprocessors'       :  %u,\\n", props.multiProcessorCount);
            printf("},\\n");
        }

        printf("]\\n");
        printf("}\\n");

        return 0;
    }
    '''  # noqa

    with _tempdir() as temp_dir:
        test_cu_path = os.path.join(temp_dir, 'test.cu')
        test_out_path = os.path.join(temp_dir, 'test.out')

        with open(test_cu_path, 'w') as f:
            f.write(code)

        proc = subprocess.Popen(
            nvcc_path + ['-o', test_out_path, test_cu_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)

        stdoutdata, stderrdata = proc.communicate()

        if proc.returncode != 0:
            raise RuntimeError("Cannot determine "
                               "compute architecture {0}"
                               .format(stderrdata))

        try:
            out = subprocess.check_output(test_out_path)
        except Exception as e:
            msg = 'Cannot execute a stub file.\nOriginal error: {0}'.format(e)
            raise Exception(msg)

        return ast.literal_eval(out)


def get_cuda_info():
    global _cuda_info

    if _cuda_info is None:
        _cuda_info = _get_cuda_info()

    return _cuda_info


def _format_cuda_version(version):
    return str(version)


def get_cuda_version(formatted=False):
    """Return CUDA Toolkit version cached in check_cuda_version()."""
    _cuda_version = get_cuda_info()['cuda_version']

    if _cuda_version < minimum_cuda_version:
        raise ValueError('CUDA version is too old: %d'
                         'CUDA v7.0 or newer is required' % _cuda_version)

    return str(_cuda_version) if formatted else _cuda_version


def get_gencode_options():
    return ["--generate-code=arch=compute_{a},code=sm_{a}".format(
            a=dev['major']*10 + dev['minor'])
            for dev in get_cuda_info()['devices']]


class _UnixCCompiler(unixccompiler.UnixCCompiler):
    src_extensions = list(unixccompiler.UnixCCompiler.src_extensions)
    src_extensions.append('.cu')

    def _compile(self, obj, src, ext, cc_args, extra_postargs, pp_opts):
        # For sources other than CUDA C ones, just call the super class method.
        if os.path.splitext(src)[1] != '.cu':
            return unixccompiler.UnixCCompiler._compile(
                self, obj, src, ext, cc_args, extra_postargs, pp_opts)

        # For CUDA C source files, compile them with NVCC.
        _compiler_so = self.compiler_so
        try:
            nvcc_path = get_nvcc_path()
            base_opts = get_compiler_base_options()
            self.set_executable('compiler_so', nvcc_path)

            cuda_version = get_cuda_version()  # noqa: triggers cuda inspection
            postargs = get_gencode_options() + [
                '-O2', '--compiler-options="-fPIC"']
            postargs += extra_postargs
            # print('NVCC options:', postargs)

            return unixccompiler.UnixCCompiler._compile(
                self, obj, src, ext, base_opts + cc_args, postargs, pp_opts)
        finally:
            self.compiler_so = _compiler_so


class _MSVCCompiler(msvccompiler.MSVCCompiler):
    _cu_extensions = ['.cu']

    src_extensions = list(unixccompiler.UnixCCompiler.src_extensions)
    src_extensions.extend(_cu_extensions)

    def _compile_cu(self, sources, output_dir=None, macros=None,
                    include_dirs=None, debug=0, extra_preargs=None,
                    extra_postargs=None, depends=None):
        # Compile CUDA C files, mainly derived from UnixCCompiler._compile().

        macros, objects, extra_postargs, pp_opts, _build = \
            self._setup_compile(output_dir, macros, include_dirs, sources,
                                depends, extra_postargs)

        compiler_so = get_nvcc_path()
        cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)
        cuda_version = get_cuda_version()  # noqa: triggers cuda inspection
        postargs = get_gencode_options() + ['-O2']
        postargs += ['-Xcompiler', '/MD']
        postargs += extra_postargs
        # print('NVCC options:', postargs)

        for obj in objects:
            try:
                src, ext = _build[obj]
            except KeyError:
                continue
            try:
                self.spawn(compiler_so + cc_args + [src, '-o', obj] + postargs)
            except errors.DistutilsExecError as e:
                raise errors.CompileError(str(e))

        return objects

    def compile(self, sources, **kwargs):
        # Split CUDA C sources and others.
        cu_sources = []
        other_sources = []
        for source in sources:
            if os.path.splitext(source)[1] == '.cu':
                cu_sources.append(source)
            else:
                other_sources.append(source)

        # Compile source files other than CUDA C ones.
        other_objects = msvccompiler.MSVCCompiler.compile(
            self, other_sources, **kwargs)

        # Compile CUDA C sources.
        cu_objects = self._compile_cu(cu_sources, **kwargs)

        # Return compiled object filenames.
        return other_objects + cu_objects


_compiler = None


def get_compiler():
    global _compiler

    if _compiler is None:
        if not PLATFORM_WIN32:
            _compiler = _UnixCCompiler()
        else:
            _compiler = _MSVCCompiler()

    return _compiler


@contextlib.contextmanager
def stdchannel_redirected(stdchannel, dest_filename):
    """
    A context manager to temporarily redirect stdout or stderr

    e.g.:

    with stdchannel_redirected(sys.stderr, os.devnull):
        if compiler.has_function('clock_gettime', libraries=['rt']):
            libraries.append('rt')
    """

    try:
        oldstdchannel = os.dup(stdchannel.fileno())
        dest_file = open(dest_filename, 'w')
        os.dup2(dest_file.fileno(), stdchannel.fileno())

        yield
    finally:
        if oldstdchannel is not None:
            os.dup2(oldstdchannel, stdchannel.fileno())
        if dest_file is not None:
            dest_file.close()


@requires_optional("cupy", cupy_import_error)
def compile_using_nvcc(source, options=None, arch=None, filename='kern.cu'):
    options = options or []

    if arch is None:
        cuda_info = get_cuda_info()
        arch = min([dev['major']*10 + dev['minor']
                    for dev in cuda_info['devices']])

    cc = get_compiler()
    settings = get_compiler_setting()
    arch = "--generate-code=arch=compute_{a},code=sm_{a}".format(a=arch)

    options += ['-cubin']

    cupy_path = resource_filename("cupy", pjoin("core", "include"))
    settings['include_dirs'].append(cupy_path)

    with _tempdir() as tmpdir:
        tmpfile = pjoin(tmpdir, filename)

        with open(tmpfile, "w") as f:
            f.write(source)

        try:
            stderr_file = pjoin(tmpdir, "stderr.txt")

            with stdchannel_redirected(sys.stderr, stderr_file):
                objects = cc.compile([tmpfile],
                                     include_dirs=settings['include_dirs'],
                                     macros=settings['define_macros'],
                                     extra_postargs=options)
        except errors.CompileError as e:
            with open(stderr_file, "r") as f:
                errs = f.read()

            lines = ["The following source code",
                     format_code(source),
                     "",
                     "created the following compilation errors",
                     "",
                     errs.strip(),
                     str(e).strip()]

            ex = errors.CompileError("\n".join(lines))
            raise (ex, None, sys.exc_info()[2])

        assert len(objects) == 1

        mod = cp.cuda.function.Module()
        mod.load_file(objects[0])

        return mod



# -*- coding: utf-8 -*-


import numpy as np
import pytest

from africanus.model.coherency import convert as np_convert
from africanus.model.coherency.cuda import convert
from africanus.model.coherency.tests.test_convert import (
                                                stokes_corr_cases,
                                                stokes_corr_int_cases,
                                                visibility_factory,
                                                vis_shape)


@pytest.mark.skip
def test_stokes_schemas(in_type, input_schema,
                        out_type, output_schema,
                        vis_shape):
    input_shape = np.asarray(input_schema).shape
    output_shape = np.asarray(output_schema).shape

    vis = visibility_factory(vis_shape, input_shape, in_type)
    xformed_vis = np_convert(vis, input_schema, output_schema)
    assert xformed_vis.shape == vis_shape + output_shape


@pytest.mark.parametrize("in_type, input_schema, out_type, output_schema",
                         stokes_corr_cases + stokes_corr_int_cases)
@pytest.mark.parametrize("vis_shape", vis_shape)
def test_cuda_convert(in_type, input_schema,
                      out_type, output_schema,
                      vis_shape):
    cp = pytest.importorskip('cupy')

    input_shape = np.asarray(input_schema).shape
    vis = visibility_factory(vis_shape, input_shape, in_type)

    cp_out = convert(cp.asarray(vis), input_schema, output_schema)
    np_out = np_convert(vis, input_schema, output_schema)

    np.testing.assert_array_almost_equal(cp.asnumpy(cp_out), np_out)



# -*- coding: utf-8 -*-


import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal
import pytest

from africanus.averaging.support import unique_time, unique_baselines
from africanus.averaging.time_and_channel_avg import time_and_channel
from africanus.averaging.time_and_channel_mapping import (row_mapper,
                                                          channel_mapper)

nchan = 16
ncorr = 4


@pytest.fixture
def time():
    return np.asarray([1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0])  # noqa


@pytest.fixture
def ant1():
    return np.asarray([0,   0,   1,   0,   0,   1,   2,   0,   0,   1],
                      dtype=np.int32)


@pytest.fixture
def ant2():
    return np.asarray([1,   2,   2,   0,   1,   2,   3,   0,   1,   2],
                      dtype=np.int32)


@pytest.fixture
def uvw():
    return np.asarray([[1.0,   1.0,  1.0],
                       [2.0,   2.0,  2.0],
                       [3.0,   3.0,  3.0],
                       [4.0,   4.0,  4.0],
                       [5.0,   5.0,  5.0],
                       [6.0,   6.0,  6.0],
                       [7.0,   7.0,  7.0],
                       [8.0,   8.0,  8.0],
                       [9.0,   9.0,  9.0],
                       [10.0, 10.0, 10.0]])


@pytest.fixture
def interval():
    data = np.asarray([1.9, 2.0, 2.1, 1.85, 1.95, 2.0, 2.05, 2.1, 2.05, 1.9])
    return 0.1 * data


@pytest.fixture
def weight(time):
    shape = (time.shape[0], ncorr)
    return np.arange(np.product(shape), dtype=np.float64).reshape(shape)


@pytest.fixture
def sigma(time):
    shape = (time.shape[0], ncorr)
    return np.arange(np.product(shape), dtype=np.float64).reshape(shape)


@pytest.fixture
def weight_spectrum(time):
    shape = (time.shape[0], nchan, ncorr)
    return np.arange(np.product(shape), dtype=np.float64).reshape(shape)


@pytest.fixture
def sigma_spectrum(time):
    shape = (time.shape[0], nchan, ncorr)
    return np.arange(np.product(shape), dtype=np.float64).reshape(shape)


@pytest.fixture
def frequency():
    return np.linspace(.856, 2*.856e9, nchan)


@pytest.fixture
def chan_width():
    return np.full(nchan, .856e9/nchan)


@pytest.fixture
def vis():
    def _vis(row, chan, fcorrs):
        flat_vis = (np.arange(row*chan*fcorrs, dtype=np.float64) +
                    np.arange(1, row*chan*fcorrs+1, dtype=np.float64)*1j)

        return flat_vis.reshape(row, chan, fcorrs)

    return _vis


@pytest.fixture
def flag():
    def _flag(row, chan, fcorrs):
        return np.random.randint(0, 2, (row, chan, fcorrs))

    return _flag


def _gen_testing_lookup(time, interval, ant1, ant2, flag_row, time_bin_secs,
                        row_meta):
    """
    Generates the same lookup as row_mapper, but different.

    Returns
    -------
    list of (float, (int, int), list of lists)
        Each tuple in the list corresponds to an output row, and
        is composed of `(avg_time, (ant1, ant2), effective_rows, nominal_rows)`

    """
    utime, _, time_inv, _ = unique_time(time)
    ubl, _, bl_inv, _ = unique_baselines(ant1, ant2)
    bl_time_lookup = np.full((ubl.shape[0], utime.shape[0]), -1,
                             dtype=np.int32)

    # Create the row index
    row_idx = np.arange(time.size)

    # Assign the row indices
    bl_time_lookup[bl_inv, time_inv] = row_idx

    # Create the time, baseline, row map
    time_bl_row_map = []

    # For each baseline, bin data that such that it fits within time_bin_secs
    # t1 - i1/2 + time_bin_secs < t2 - i2/2
    # where (t1, t2) and (i1, i2) are the times and intervals associated
    # with two different samples in the baseline.
    # Compute two different bins
    # 1. Effective row bin, which only includes unflagged rows
    #    unless the entire bin is flagged, in which case it includes flagged
    #    data
    # 2. Nominal row bin, which includes both flagged and unflagged rows

    for bl, (a1, a2) in enumerate(ubl):
        bl_row_idx = bl_time_lookup[bl, :]

        effective_bin_map = []
        effective_map = []

        nominal_bin_map = []
        nominal_map = []

        for ri in bl_row_idx:
            if ri == -1:
                continue

            half_int = 0.5 * interval[ri]

            # We're starting a new bin
            if len(nominal_map) == 0:
                bin_low = time[ri] - half_int
            # Reached passed the endpoint of the bin, start a new one
            elif time[ri] + half_int - bin_low > time_bin_secs:
                if len(effective_map) > 0:
                    effective_bin_map.append(effective_map)
                    nominal_bin_map.append(nominal_map)
                # No effective samples, the entire bin must be flagged
                elif len(nominal_map) > 0:
                    effective_bin_map.append(nominal_map)
                    nominal_bin_map.append(nominal_map)
                else:
                    raise ValueError("Zero-filled bin")

                effective_map = []
                nominal_map = []

            # Effective only includes unflagged samples
            if flag_row[ri] == 0:
                effective_map.append(ri)

            # Nominal includes all samples
            nominal_map.append(ri)

        # Add any remaining values
        if len(effective_map) > 0:
            effective_bin_map.append(effective_map)
            nominal_bin_map.append(nominal_map)
        # No effective samples, the entire bin must be flagged
        # so we add nominal samples
        elif len(nominal_map) > 0:
            effective_bin_map.append(nominal_map)
            nominal_bin_map.append(nominal_map)

        # Produce a (avg_time, bl, effective_rows, nominal_rows) tuple
        time_bl_row_map.extend((time[nrows].mean(), (a1, a2), erows, nrows)
                               for erows, nrows
                               in zip(effective_bin_map, nominal_bin_map))

    # Sort lookup sorted on averaged times
    return sorted(time_bl_row_map, key=lambda tup: tup[0])


def _calc_sigma(sigma, weight, idx):
    sigma = sigma[idx]
    weight = weight[idx]

    numerator = (sigma**2 * weight**2).sum(axis=0)
    denominator = weight.sum(axis=0)**2
    denominator[denominator == 0.0] = 1.0

    return np.sqrt(numerator / denominator)


@pytest.mark.parametrize("flagged_rows", [
    [], [8, 9], [4], [0, 1],
])
@pytest.mark.parametrize("time_bin_secs", [1, 2, 3, 4])
@pytest.mark.parametrize("chan_bin_size", [1, 3, 5])
def test_averager(time, ant1, ant2, flagged_rows,
                  uvw, interval, weight, sigma,
                  frequency, chan_width,
                  vis, flag,
                  weight_spectrum, sigma_spectrum,
                  time_bin_secs, chan_bin_size):

    time_centroid = time
    exposure = interval

    vis = vis(time.shape[0], nchan, ncorr)
    flag = flag(time.shape[0], nchan, ncorr)

    flag_row = np.zeros(time.shape, dtype=np.uint8)

    # flagged_row and flag should agree
    flag_row[flagged_rows] = 1
    flag[flagged_rows, :, :] = 1

    row_meta = row_mapper(time, interval, ant1, ant2, flag_row, time_bin_secs)
    chan_map, chan_bins = channel_mapper(nchan, chan_bin_size)

    time_bl_row_map = _gen_testing_lookup(time_centroid, exposure, ant1, ant2,
                                          flag_row, time_bin_secs,
                                          row_meta)

    # Effective and Nominal rows associated with each output row
    eff_idx, nom_idx = zip(*[(nrows, erows) for _, _, nrows, erows
                             in time_bl_row_map])

    eff_idx = [ei for ei in eff_idx if len(ei) > 0]

    # Check that the averaged times from the test and accelerated lookup match
    assert_array_equal([t for t, _, _, _ in time_bl_row_map],
                       row_meta.time)

    avg = time_and_channel(time, interval, ant1, ant2,
                           flag_row=flag_row,
                           time_centroid=time, exposure=exposure, uvw=uvw,
                           weight=weight, sigma=sigma,
                           chan_freq=frequency, chan_width=chan_width,
                           vis=vis, flag=flag,
                           weight_spectrum=weight_spectrum,
                           sigma_spectrum=sigma_spectrum,
                           time_bin_secs=time_bin_secs,
                           chan_bin_size=chan_bin_size)

    # Take mean time, but first ant1 and ant2
    expected_time_centroids = [time_centroid[i].mean(axis=0) for i in eff_idx]
    expected_times = [time[i].mean(axis=0) for i in nom_idx]
    expected_ant1 = [ant1[i[0]] for i in nom_idx]
    expected_ant2 = [ant2[i[0]] for i in nom_idx]
    expected_flag_row = [flag_row[i].any(axis=0) for i in eff_idx]

    # Take mean average, but sum of interval and exposure
    expected_uvw = [uvw[i].mean(axis=0) for i in eff_idx]
    expected_interval = [interval[i].sum(axis=0) for i in nom_idx]
    expected_exposure = [exposure[i].sum(axis=0) for i in eff_idx]
    expected_weight = [weight[i].sum(axis=0) for i in eff_idx]
    expected_sigma = [_calc_sigma(sigma, weight, i) for i in eff_idx]

    assert_array_equal(row_meta.time, expected_times)
    assert_array_equal(row_meta.interval, expected_interval)
    assert_array_equal(row_meta.flag_row, expected_flag_row)
    assert_array_equal(avg.antenna1, expected_ant1)
    assert_array_equal(avg.antenna2, expected_ant2)
    assert_array_equal(avg.time_centroid, expected_time_centroids)
    assert_array_equal(avg.exposure, expected_exposure)
    assert_array_equal(avg.uvw, expected_uvw)
    assert_array_equal(avg.weight, expected_weight)
    assert_array_equal(avg.sigma, expected_sigma)

    chan_avg_shape = (row_meta.interval.shape[0], chan_bins, flag.shape[2])

    assert avg.vis.shape == chan_avg_shape
    assert avg.flag.shape == chan_avg_shape
    assert avg.weight_spectrum.shape == chan_avg_shape
    assert avg.sigma_spectrum.shape == chan_avg_shape

    chan_ranges = np.nonzero(np.ediff1d(chan_map, to_begin=1, to_end=1))[0]

    # Three python loops. Slow, but works...
    # Figure out some way to remove loops with numpy
    for orow, idx in enumerate(eff_idx):
        for ch, (cs, ce) in enumerate(zip(chan_ranges[:-1], chan_ranges[1:])):
            for corr in range(ncorr):
                chunk_flags = flag[idx, cs:ce, corr] == 1
                expected_flags = chunk_flags.all()
                assert_array_equal(expected_flags,  avg.flag[orow, ch, corr])

                exp_vis = vis[idx, cs:ce, corr]
                exp_wts = weight_spectrum[idx, cs:ce, corr]
                exp_sigma = sigma_spectrum[idx, cs:ce, corr]

                # If all rows+chans in this correlation are flagged
                # then we use flagged values to calculate vis
                if expected_flags:
                    exp_vis = exp_vis[chunk_flags]
                    exp_wts = exp_wts[chunk_flags]
                    exp_sigma = exp_sigma[chunk_flags]
                # Otherwise we used unflagged values to calculate vis
                else:
                    exp_vis = exp_vis[~chunk_flags]
                    exp_wts = exp_wts[~chunk_flags]
                    exp_sigma = exp_sigma[~chunk_flags]

                exp_vis = (exp_vis*exp_wts).sum()
                exp_sigma = (exp_sigma**2 * exp_wts**2).sum()
                exp_wts = exp_wts.sum()

                if exp_wts != 0.0:
                    exp_vis = exp_vis / exp_wts
                    exp_sigma = np.sqrt(exp_sigma / (exp_wts**2))

                assert_array_almost_equal(exp_vis, avg.vis[orow, ch, corr])
                assert_array_almost_equal(exp_sigma,
                                          avg.sigma_spectrum[orow, ch, corr])
                assert_array_almost_equal(exp_wts,
                                          avg.weight_spectrum[orow, ch, corr])


@pytest.mark.parametrize("flagged_rows", [
    [], [8, 9], [4], [0, 1],
])
@pytest.mark.parametrize("time_bin_secs", [1, 2, 3, 4])
@pytest.mark.parametrize("chan_bin_size", [1, 3, 5])
def test_dask_averager(time, ant1, ant2, flagged_rows,
                       uvw, interval, weight, sigma,
                       frequency, chan_width,
                       vis, flag,
                       weight_spectrum, sigma_spectrum,
                       time_bin_secs, chan_bin_size):

    da = pytest.importorskip('dask.array')

    from africanus.averaging.dask import time_and_channel as dask_avg

    rc = (6, 4)
    fc = (4, 4, 4, 4)
    cc = (4,)

    rows = sum(rc)
    chans = sum(fc)
    corrs = sum(cc)

    time_centroid = time
    exposure = interval

    vis = vis(rows, chans, corrs)
    flag = flag(rows, chans, corrs)
    flag_row = np.zeros(time_centroid.shape[0], dtype=np.uint8)
    flag_row[flagged_rows] = 1
    flag[flagged_rows, :, :] = 1

    np_avg = time_and_channel(time_centroid, exposure, ant1, ant2,
                              flag_row=flag_row,
                              chan_freq=frequency, chan_width=chan_width,
                              effective_bw=chan_width, resolution=chan_width,
                              vis=vis, flag=flag,
                              weight_spectrum=weight_spectrum,
                              time_bin_secs=time_bin_secs,
                              chan_bin_size=chan_bin_size)

    # Using chunks == shape, the dask version should match the numpy version
    da_time_centroid = da.from_array(time_centroid, chunks=rows)
    da_exposure = da.from_array(exposure, chunks=rows)
    da_flag_row = da.from_array(flag_row, chunks=rows)
    da_weight = da.from_array(weight, chunks=(rows, corrs))
    da_sigma = da.from_array(sigma, chunks=(rows, corrs))
    da_ant1 = da.from_array(ant1, chunks=rows)
    da_ant2 = da.from_array(ant2, chunks=rows)
    da_chan_freq = da.from_array(frequency, chunks=chans)
    da_chan_width = da.from_array(chan_width, chunks=chans)
    da_weight_spectrum = da.from_array(weight_spectrum,
                                       chunks=(rows, chans, corrs))
    da_vis = da.from_array(vis, chunks=(rows, chans, corrs))
    da_flag = da.from_array(flag, chunks=(rows, chans, corrs))

    avg = dask_avg(da_time_centroid, da_exposure, da_ant1, da_ant2,
                   flag_row=da_flag_row,
                   chan_freq=da_chan_freq, chan_width=da_chan_width,
                   effective_bw=da_chan_width, resolution=da_chan_width,
                   weight=da_weight, sigma=da_sigma,
                   vis=da_vis, flag=da_flag,
                   weight_spectrum=da_weight_spectrum,
                   time_bin_secs=time_bin_secs,
                   chan_bin_size=chan_bin_size)

    # Compute all the averages in one go
    (avg_time_centroid, avg_exposure, avg_flag_row,
     avg_chan_freq, avg_chan_width,
     avg_resolution, avg_vis, avg_flag) = da.compute(
                              avg.time_centroid,
                              avg.exposure,
                              avg.flag_row,
                              avg.chan_freq,
                              avg.chan_width,
                              avg.resolution,
                              avg.vis, avg.flag)

    # Should match
    assert_array_equal(np_avg.time_centroid, avg_time_centroid)
    assert_array_equal(np_avg.exposure, avg_exposure)
    assert_array_equal(np_avg.flag_row, avg_flag_row)
    assert_array_equal(np_avg.vis, avg_vis)
    assert_array_equal(np_avg.flag, avg_flag)
    assert_array_equal(np_avg.chan_freq, avg_chan_freq)
    assert_array_equal(np_avg.chan_width, avg_chan_width)
    assert_array_equal(np_avg.resolution, avg_resolution)

    # We can average chunked arrays too, but these will not necessarily
    # match the numpy version
    da_time_centroid = da.from_array(time_centroid, chunks=(rc,))
    da_exposure = da.from_array(exposure, chunks=(rc,))
    da_flag_row = da.from_array(flag_row, chunks=(rc,))
    da_ant1 = da.from_array(ant1, chunks=(rc,))
    da_ant2 = da.from_array(ant2, chunks=(rc,))
    da_chan_freq = da.from_array(frequency, chunks=(fc,))
    da_chan_width = da.from_array(chan_width, chunks=(fc,))
    da_flag = da.from_array(flag, chunks=(rc, fc, cc))
    da_vis = da.from_array(vis, chunks=(rc, fc, cc))
    da_flag = da.from_array(flag, chunks=(rc, fc, cc))

    avg = dask_avg(da_time_centroid, da_exposure, da_ant1, da_ant2,
                   flag_row=da_flag_row,
                   chan_freq=da_chan_freq, chan_width=da_chan_width,
                   vis=da_vis, flag=da_flag,
                   time_bin_secs=time_bin_secs,
                   chan_bin_size=chan_bin_size)

    # Compute all the fields
    fields = [getattr(avg, f) for f in avg._fields]
    avg = type(avg)(*da.compute(fields)[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import argparse
import logging

from africanus.filters import convolution_filter
from africanus.util.cmdline import parse_python_assigns
from africanus.util.requirements import requires_optional

import numpy as np

try:
    import matplotlib.pyplot as plt
    from matplotlib import cm
    from matplotlib.ticker import LinearLocator, FormatStrFormatter
    from mpl_toolkits.mplot3d import Axes3D  # noqa
except ImportError as e:
    mpl_ie = e
else:
    mpl_ie = None


def create_parser():
    p = argparse.ArgumentParser()
    p.add_argument("filter", choices=['kaiser-bessel', 'sinc'],
                   default='kaiser-bessel')
    p.add_argument("-hs", "--half-support", default=3, type=int)
    p.add_argument("-os", "--oversample", default=63, type=int)
    p.add_argument("-n", "--normalise", dest="normalise", action="store_true",
                   help="Normalise filter by it's volume")
    p.add_argument("--no-normalise", dest="normalise", action="store_false",
                   help="Do not normalise the filter by it's volume")
    p.add_argument("-k", "--kwargs", default="", type=parse_python_assigns,
                   help="Extra keywords arguments used to create the filter. "
                        "For example 'beta=2.3' to specify a beta shape "
                        "parameter for the Kaiser Bessel")

    return p


@requires_optional('matplotlib.pyplot', 'mpl_toolkits.mplot3d', mpl_ie)
def _plot_filter(data):
    X, Y = np.mgrid[-1:1:1j*data.shape[0], -1:1:1j*data.shape[1]]

    fig = plt.figure()
    ax = fig.gca(projection='3d')

    # Plot the surface.
    surf = ax.plot_surface(X, Y, data, cmap=cm.coolwarm,
                           linewidth=0, antialiased=False)

    zmax = data.max()

    ax.plot([0, 0], [0, 0], zs=[0, 1.2*zmax], color='black')
    ax.plot([0, 0], [-1, 1], zs=[zmax, zmax], color='black')
    ax.plot([-1, 1], [0, 0], zs=[zmax, zmax], color='black')

    # Customize the z axis.
    ax.zaxis.set_major_locator(LinearLocator(10))
    ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))

    # Add a color bar which maps values to colors.
    fig.colorbar(surf, shrink=0.5, aspect=5)

    plt.show()


def main():
    logging.basicConfig(level=logging.INFO, format="%(message)s")

    args = create_parser().parse_args()

    logging.info("Creating %s filter with half support of %d "
                 "and %d oversampling" %
                 (args.filter, args.half_support, args.oversample))

    if len(args.kwargs) > 0:
        logging.info("Extra keywords %s" % args.kwargs)

    conv_filter = convolution_filter(args.half_support,
                                     args.oversample,
                                     args.filter,
                                     normalise=args.normalise,
                                     **args.kwargs)

    _plot_filter(np.abs(conv_filter.filter_taps))



# -*- coding: utf-8 -*-


import numpy as np

from africanus.util.docs import DocstringTemplate
from africanus.util.numba import generated_jit
from africanus.constants import c as lightspeed


@generated_jit(nopython=True, nogil=True, cache=True)
def gaussian(uvw, frequency, shape_params):
    # https://en.wikipedia.org/wiki/Full_width_at_half_maximum
    fwhm = 2.0 * np.sqrt(2.0 * np.log(2.0))
    fwhminv = 1.0 / fwhm
    gauss_scale = fwhminv * np.sqrt(2.0) * np.pi / lightspeed

    dtype = np.result_type(*(np.dtype(a.dtype.name) for
                             a in (uvw, frequency, shape_params)))

    def impl(uvw, frequency, shape_params):
        nsrc = shape_params.shape[0]
        nrow = uvw.shape[0]
        nchan = frequency.shape[0]

        shape = np.empty((nsrc, nrow, nchan), dtype=dtype)
        scaled_freq = np.empty_like(frequency)

        # Scale each frequency
        for f in range(frequency.shape[0]):
            scaled_freq[f] = frequency[f] * gauss_scale

        for s in range(shape_params.shape[0]):
            emaj, emin, angle = shape_params[s]

            # Convert to l-projection, m-projection, ratio
            el = emaj * np.sin(angle)
            em = emaj * np.cos(angle)
            er = emin / (1.0 if emaj == 0.0 else emaj)

            for r in range(uvw.shape[0]):
                u, v, w = uvw[r]

                u1 = (u*em - v*el)*er
                v1 = u*el + v*em

                for f in range(scaled_freq.shape[0]):
                    fu1 = u1*scaled_freq[f]
                    fv1 = v1*scaled_freq[f]

                    shape[s, r, f] = np.exp(-(fu1*fu1 + fv1*fv1))

        return shape

    return impl


GAUSSIAN_DOCS = DocstringTemplate(r"""
Computes the Gaussian Shape Function.

.. math::

    & \lambda^\prime = 2 \lambda \pi \\
    & r = \frac{e_{min}}{e_{maj}} \\
    & u_{1} = (u \, e_{maj} \, cos(\alpha) - v \, e_{maj} \, sin(\alpha))
      r \lambda^\prime \\
    & v_{1} = (u \, e_{maj} \, sin(\alpha) - v \, e_{maj} \, cos(\alpha))
      \lambda^\prime \\
    & \textrm{shape} = e^{(-u_{1}^2 - v_{1}^2)}

where:

- :math:`u` and :math:`v` are the UV coordinates and
  :math:`\lambda` the frequency.
- :math:`e_{maj}` and :math:`e_{min}` are the major and minor axes
  and :math:`\alpha` the position angle.

Parameters
----------
uvw : $(array_type)
    UVW coordinates of shape :code:`(row, 3)`
frequency : $(array_type)
    frequencies of shape :code:`(chan,)`
shape_param : $(array_type)
    Gaussian Shape Parameters of shape :code:`(source, 3)`
    where the second dimension contains the
    `(emajor, eminor, angle)` parameters describing
    the shape of the Gaussian

Returns
-------
gauss_shape : $(array_type)
    Shape parameters of shape :code:`(source, row, chan)`
""")

try:
    gaussian.__doc__ = GAUSSIAN_DOCS.substitute(
                            array_type=":class:`numpy.ndarray`")
except KeyError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import math

import numpy as np

from africanus.util.numba import jit


@jit(nopython=True, nogil=True, cache=True)
def _nb_transform_sources(lm, parallactic_angles, pointing_errors,
                          antenna_scaling, frequency, coords):
    """
    numba implementation of
    :func:`~africanus.rime.transform_sources`
    """
    _, nsrc, ntime, na, nchan = coords.shape

    for t in range(ntime):
        for a in range(na):
            pa_sin = math.sin(parallactic_angles[t, a])
            pa_cos = math.cos(parallactic_angles[t, a])

            for s in range(nsrc):
                l, m = lm[s]

                # Rotate source coordinate by parallactic angle
                l = l*pa_cos - m*pa_sin  # noqa
                m = l*pa_sin + m*pa_cos

                # Add pointing errors
                l += pointing_errors[t, a, 0]  # noqa
                m += pointing_errors[t, a, 1]

                # Scale by antenna scaling factors
                for c in range(nchan):
                    coords[0, s, t, a, c] = l*antenna_scaling[a, c]
                    coords[1, s, t, a, c] = m*antenna_scaling[a, c]
                    coords[2, s, t, a, c] = frequency[c]

    return coords


def transform_sources(lm, parallactic_angles, pointing_errors,
                      antenna_scaling, frequency, dtype=None):
    """
    Creates beam sampling coordinates suitable for use
    in :func:`~africanus.rime.beam_cube_dde` by:

    1. Rotating ``lm`` coordinates by the ``parallactic_angles``
    2. Adding ``pointing_errors``
    3. Scaling by ``antenna_scaling``

    Parameters
    ----------
    lm : :class:`numpy.ndarray`
        LM coordinates of shape :code:`(src,2)` in radians
        offset from the phase centre.
    parallactic_angles : :class:`numpy.ndarray`
        parallactic angles of shape :code:`(time, antenna)`
        in radians.
    pointing_errors : :class:`numpy.ndarray`
        LM pointing errors for each antenna at
        each timestep in radians.
        Has shape :code:`(time, antenna, 2)`
    antenna_scaling : :class:`numpy.ndarray`
        antenna scaling factor for each channel and
        each antenna. Has shape :code:`(antenna, chan)`
    frequency : :class:`numpy.ndarray`
        frequencies for each channel. Has shape :code:`(chan,)`
    dtype : :class:`numpy.dtype`, optional
        Numpy dtype of result array. Should be float32 or float64.
        Defaults to float64


    Returns
    -------
    coords : :class:`numpy.ndarray`
        coordinates of shape :code:`(3, src, time, antenna, chan)`
        where each coordinate component represents **l**, **m** and
        **frequency**, respectively.
    """

    ntime, na = parallactic_angles.shape
    nsrc = lm.shape[0]
    assert (ntime, na, 2) == pointing_errors.shape
    nchan = antenna_scaling.shape[1]
    assert nchan == frequency.shape[0]

    dtype = np.float64 if dtype is None else dtype
    coords = np.empty((3, nsrc, ntime, na, nchan), dtype=dtype)

    return _nb_transform_sources(lm, parallactic_angles, pointing_errors,
                                 antenna_scaling, frequency, coords)



# -*- coding: utf-8 -*-


from operator import getitem

from africanus.averaging.time_and_channel_mapping import (
                row_mapper as np_row_mapper,
                channel_mapper as np_channel_mapper)
from africanus.averaging.time_and_channel_avg import (
                row_average as np_row_average,
                row_chan_average as np_row_chan_average,
                chan_average as np_chan_average,
                merge_flags as np_merge_flags,
                AVERAGING_DOCS,
                AverageOutput, ChannelAverageOutput,
                RowAverageOutput, RowChanAverageOutput)

from africanus.util.requirements import requires_optional

import numpy as np

try:
    from dask.base import tokenize
    import dask.array as da
    import dask.blockwise as db
    from dask.highlevelgraph import HighLevelGraph
except ImportError as e:
    dask_import_error = e
else:
    dask_import_error = None


def chan_metadata(row_chan_arrays, chan_arrays, chan_bin_size):
    """ Create dask array with channel metadata for each chunk channel """
    chan_chunks = None

    for array in row_chan_arrays:
        if array is not None:
            chan_chunks = array.chunks[1]
            break

    if chan_chunks is None:
        for array in chan_arrays:
            if array is not None:
                chan_chunks = array.chunks[0]
                break

    if chan_chunks is None:
        return None

    # Create a dask channel mapping structure
    name = "channel-mapper-" + tokenize(chan_chunks, chan_bin_size)
    layers = {(name, i): (np_channel_mapper, c, chan_bin_size)
              for i, c in enumerate(chan_chunks)}
    graph = HighLevelGraph.from_collections(name, layers, ())
    chunks = (chan_chunks,)
    chan_mapper = da.Array(graph, name, chunks, dtype=np.object)

    return chan_mapper


def row_mapper(time, interval, antenna1, antenna2,
               flag_row=None, time_bin_secs=1.0):
    """ Create a dask row mapping structure for each row chunk """
    return da.blockwise(np_row_mapper, ("row",),
                        time, ("row",),
                        interval, ("row",),
                        antenna1, ("row",),
                        antenna2, ("row",),
                        flag_row, None if flag_row is None else ("row",),
                        adjust_chunks={"row": lambda x: np.nan},
                        time_bin_secs=time_bin_secs,
                        meta=np.empty((0,), dtype=np.object),
                        dtype=np.object)


def _getitem_row(avg, idx, array, dims):
    """ Extract row-like arrays from a dask array of tuples """
    assert dims[0] == "row"

    name = ("row-average-getitem-%d-" % idx) + tokenize(avg, idx)
    layers = db.blockwise(getitem, name, dims,
                          avg.name, ("row",),
                          idx, None,
                          new_axes=dict(zip(dims[1:], array.shape[1:])),
                          numblocks={avg.name: avg.numblocks})
    graph = HighLevelGraph.from_collections(name, layers, (avg,))
    chunks = avg.chunks + tuple((s,) for s in array.shape[1:])

    return da.Array(graph, name, chunks,
                    meta=np.empty((0,)*len(dims), dtype=array.dtype),
                    dtype=array.dtype)


def _row_average_wrapper(row_meta, ant1, ant2, flag_row,
                         time_centroid, exposure, uvw,
                         weight, sigma):
    return np_row_average(row_meta, ant1, ant2, flag_row,
                          time_centroid, exposure,
                          uvw[0] if uvw is not None else None,
                          weight[0] if weight is not None else None,
                          sigma[0] if sigma is not None else None)


def row_average(row_meta, ant1, ant2, flag_row=None,
                time_centroid=None, exposure=None, uvw=None,
                weight=None, sigma=None):
    """ Average row-based dask arrays """

    rd = ("row",)
    rcd = ("row", "corr")

    # (output, array, dims)
    args = [(False, row_meta, rd),
            (True, ant1, rd),
            (True, ant2, rd),
            (False, flag_row, None if flag_row is None else rd),
            (True, time_centroid, None if time_centroid is None else rd),
            (True, exposure, None if exposure is None else rd),
            (True, uvw, None if uvw is None else ("row", "uvw")),
            (True, weight, None if weight is None else rcd),
            (True, sigma, None if sigma is None else rcd)]

    avg = da.blockwise(_row_average_wrapper, rd,
                       *(v for pair in args for v in pair[1:]),
                       align_arrays=False,
                       adjust_chunks={"row": lambda x: np.nan},
                       meta=np.empty((0,)*len(rd), dtype=np.object),
                       dtype=np.object)

    # ant1, ant2, time_centroid, exposure, uvw, weight, sigma
    out_args = [(a, dims) for out, a, dims in args if out is True]

    tuple_gets = [None if a is None else _getitem_row(avg, i, a, dims)
                  for i, (a, dims) in enumerate(out_args)]

    return RowAverageOutput(*tuple_gets)


def _getitem_row_chan(avg, idx, dtype):
    """ Extract (row,chan,corr) arrays from dask array of tuples """
    name = ("row-chan-average-getitem-%d-" % idx) + tokenize(avg, idx)
    dim = ("row", "chan", "corr")

    layers = db.blockwise(getitem, name, dim,
                          avg.name, dim,
                          idx, None,
                          numblocks={avg.name: avg.numblocks})

    graph = HighLevelGraph.from_collections(name, layers, (avg,))
    return da.Array(graph, name, avg.chunks,
                    meta=np.empty((0,)*len(dim), dtype=np.object),
                    dtype=dtype)


_row_chan_avg_dims = ("row", "chan", "corr")


def row_chan_average(row_meta, chan_meta, flag_row=None, weight=None,
                     vis=None, flag=None,
                     weight_spectrum=None, sigma_spectrum=None,
                     chan_bin_size=1):
    """ Average (row,chan,corr)-based dask arrays """

    if chan_meta is None:
        return RowChanAverageOutput(None, None, None, None)

    # We don't know how many rows are in each row chunk,
    # but we can simply divide each channel chunk size by the bin size
    adjust_chunks = {
        "row": lambda r: np.nan,
        "chan": lambda c: (c + chan_bin_size - 1) // chan_bin_size
    }

    flag_row_dims = None if flag_row is None else ("row",)
    weight_dims = None if weight is None else ("row", "corr")
    vis_dims = None if vis is None else _row_chan_avg_dims
    flag_dims = None if flag is None else _row_chan_avg_dims
    ws_dims = None if weight_spectrum is None else _row_chan_avg_dims
    ss_dims = None if sigma_spectrum is None else _row_chan_avg_dims

    avg = da.blockwise(np_row_chan_average, _row_chan_avg_dims,
                       row_meta, ("row",),
                       chan_meta, ("chan",),
                       flag_row, flag_row_dims,
                       weight, weight_dims,
                       vis, vis_dims,
                       flag, flag_dims,
                       weight_spectrum, ws_dims,
                       sigma_spectrum, ss_dims,
                       align_arrays=False,
                       adjust_chunks=adjust_chunks,
                       meta=np.empty((0,)*len(_row_chan_avg_dims),
                                     dtype=np.object),
                       dtype=np.object)

    tuple_gets = (None if a is None else _getitem_row_chan(avg, i, a.dtype)
                  for i, a in enumerate([vis, flag,
                                         weight_spectrum,
                                         sigma_spectrum]))

    return RowChanAverageOutput(*tuple_gets)


def _getitem_chan(avg, idx, dtype):
    """ Extract row-like arrays from a dask array of tuples """
    name = ("chan-average-getitem-%d-" % idx) + tokenize(avg, idx)
    layers = db.blockwise(getitem, name, ("chan",),
                          avg.name, ("chan",),
                          idx, None,
                          numblocks={avg.name: avg.numblocks})
    graph = HighLevelGraph.from_collections(name, layers, (avg,))
    return da.Array(graph, name, avg.chunks,
                    meta=np.empty((0,), dtype=dtype),
                    dtype=dtype)


def chan_average(chan_meta, chan_freq=None, chan_width=None,
                 effective_bw=None, resolution=None, chan_bin_size=1):

    if chan_meta is None:
        return ChannelAverageOutput(None, None)

    adjust_chunks = {
        "chan": lambda c: (c + chan_bin_size - 1) // chan_bin_size
    }

    cdim = ("chan",)

    avg = da.blockwise(np_chan_average, cdim,
                       chan_meta, cdim,
                       chan_freq, None if chan_freq is None else cdim,
                       chan_width, None if chan_width is None else cdim,
                       effective_bw, None if effective_bw is None else cdim,
                       resolution, None if resolution is None else cdim,
                       adjust_chunks=adjust_chunks,
                       meta=np.empty((0,), dtype=np.object),
                       dtype=np.object)

    tuple_gets = (None if a is None else _getitem_chan(avg, i, a.dtype)
                  for i, a in enumerate([chan_freq, chan_width,
                                         effective_bw, resolution]))

    return ChannelAverageOutput(*tuple_gets)


def merge_flags(flag_row, flag):
    """ Perform flag merging on dask arrays """
    if flag_row is None and flag is not None:
        return da.blockwise(np_merge_flags, "r",
                            flag_row, None,
                            flag, "rfc",
                            concatenate=True,
                            dtype=flag.dtype)
    elif flag_row is not None and flag is None:
        return da.blockwise(np_merge_flags, "r",
                            flag_row, "r",
                            None, None,
                            dtype=flag_row.dtype)
    elif flag_row is not None and flag is not None:
        return da.blockwise(np_merge_flags, "r",
                            flag_row, "r",
                            flag, "rfc",
                            concatenate=True,
                            dtype=flag_row.dtype)
    else:
        return None


@requires_optional("dask.array", dask_import_error)
def time_and_channel(time, interval, antenna1, antenna2,
                     time_centroid=None, exposure=None, flag_row=None,
                     uvw=None, weight=None, sigma=None,
                     chan_freq=None, chan_width=None,
                     effective_bw=None, resolution=None,
                     vis=None, flag=None,
                     weight_spectrum=None, sigma_spectrum=None,
                     time_bin_secs=1.0, chan_bin_size=1):

    row_chan_arrays = (vis, flag, weight_spectrum, sigma_spectrum)
    chan_arrays = (chan_freq, chan_width, effective_bw, resolution)

    # The flow of this function should match that of the numba
    # time_and_channel implementation

    # Merge flag_row and flag arrays
    flag_row = merge_flags(flag_row, flag)

    # Generate row mapping metadata
    row_meta = row_mapper(time, interval,
                          antenna1, antenna2,
                          flag_row=flag_row,
                          time_bin_secs=time_bin_secs)

    # Generate channel mapping metadata
    chan_meta = chan_metadata(row_chan_arrays, chan_arrays, chan_bin_size)

    # Average row data
    row_data = row_average(row_meta, antenna1, antenna2,
                           flag_row=flag_row,
                           time_centroid=time_centroid,
                           exposure=exposure, uvw=uvw,
                           weight=weight, sigma=sigma)

    # Average channel data
    row_chan_data = row_chan_average(row_meta, chan_meta,
                                     flag_row=flag_row, weight=weight,
                                     vis=vis, flag=flag,
                                     weight_spectrum=weight_spectrum,
                                     sigma_spectrum=sigma_spectrum,
                                     chan_bin_size=chan_bin_size)

    chan_data = chan_average(chan_meta,
                             chan_freq=chan_freq,
                             chan_width=chan_width,
                             effective_bw=effective_bw,
                             resolution=resolution)

    # Merge output tuples
    return AverageOutput(_getitem_row(row_meta, 1, time, ("row",)),
                         _getitem_row(row_meta, 2, interval, ("row",)),
                         (_getitem_row(row_meta, 3, flag_row, ("row",))
                          if flag_row is not None else None),
                         row_data.antenna1,
                         row_data.antenna2,
                         row_data.time_centroid,
                         row_data.exposure,
                         row_data.uvw,
                         row_data.weight,
                         row_data.sigma,
                         chan_data.chan_freq,
                         chan_data.chan_width,
                         chan_data.effective_bw,
                         chan_data.resolution,
                         row_chan_data.vis,
                         row_chan_data.flag,
                         row_chan_data.weight_spectrum,
                         row_chan_data.sigma_spectrum)


try:
    time_and_channel.__doc__ = AVERAGING_DOCS.substitute(
                                    array_type=":class:`dask.array.Array`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


from africanus.model.coherency.conversion import (convert_setup,
                                                  convert_impl,
                                                  CONVERT_DOCS)

from africanus.util.requirements import requires_optional

try:
    import dask.array as da
except ImportError as e:
    da_import_error = e
else:
    da_import_error = None


def _wrapper(np_input, mapping=None, in_shape=None,
             out_shape=None, dtype_=None):
    result = convert_impl(np_input, mapping, in_shape,
                          out_shape, dtype_)

    # Introduce extra singleton dimension at the end of our shape
    return result.reshape(result.shape + (1,) * len(in_shape))


@requires_optional("dask.array", da_import_error)
def convert(input, input_schema, output_schema):
    mapping, in_shape, out_shape, dtype = convert_setup(input,
                                                        input_schema,
                                                        output_schema)

    n_free_dims = len(input.shape) - len(in_shape)
    free_dims = tuple("dim-%d" % i for i in range(n_free_dims))
    in_corr_dims = tuple("icorr-%d" % i for i in range(len(in_shape)))
    out_corr_dims = tuple("ocorr-%d" % i for i in range(len(out_shape)))

    # Output dimension are new dimensions
    new_axes = {d: s for d, s in zip(out_corr_dims, out_shape)}

    # Note the dummy in_corr_dims introduced at the end of our output,
    # We do this to prevent a contraction over the input dimensions
    # (which can be arbitrary) within the wrapper class
    res = da.core.blockwise(_wrapper, free_dims + out_corr_dims + in_corr_dims,
                            input, free_dims + in_corr_dims,
                            mapping=mapping,
                            in_shape=in_shape,
                            out_shape=out_shape,
                            new_axes=new_axes,
                            dtype_=dtype,
                            dtype=dtype)

    # Now contract over the dummy dimensions
    start = len(free_dims) + len(out_corr_dims)
    end = start + len(in_corr_dims)
    return res.sum(axis=list(range(start, end)))


try:
    convert.__doc__ = CONVERT_DOCS.substitute(
                                array_type=":class:`dask.array.Array`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


from ast import literal_eval
import math
import re

hour_re = re.compile(r"(?P<sign>[+-]*)"
                     r"(?P<hours>\d+):"
                     r"(?P<mins>\d+):"
                     r"(?P<secs>\d+\.?\d*)")

deg_re = re.compile(r"(?P<sign>[+-])*"
                    r"(?P<degs>\d+)\."
                    r"(?P<mins>\d+)\."
                    r"(?P<secs>\d+\.?\d*)")


def _hour_converter(hour_str):
    m = hour_re.match(hour_str)

    if not m:
        raise ValueError("Error parsing '%s'" % hour_str)

    value = float(m.group("hours")) / 24.0
    value += float(m.group("mins")) / (24.0*60.0)
    value += float(m.group("secs")) / (24.0*60.0*60.0)

    if m.group("sign") == '-':
        value = -value

    return 2.0 * math.pi * value


def _deg_converter(deg_str):
    m = deg_re.match(deg_str)

    if not m:
        raise ValueError("Error parsing '%s'" % deg_str)

    value = float(m.group("degs")) / 360.0
    value += float(m.group("mins")) / (360.0*60.0)
    value += float(m.group("secs")) / (360.0*60.0*60.0)

    if m.group("sign") == '-':
        value = -value

    return 2.0 * math.pi * value


_COLUMN_CONVERTERS = {
    'Name': str,
    'Type': str,
    'Ra': _hour_converter,
    'Dec': _deg_converter,
    'I': float,
    'SpectralIndex': literal_eval,
    'LogarithmicSI': lambda x: bool(x == "true"),
    'ReferenceFrequency': float,
    'MajorAxis': float,
    'MinorAxis': float,
    'Orientation': float,
}


# Split on commas, ignoring within [] brackets
_COMMA_SPLIT_RE = re.compile(r',\s*(?=[^\]]*(?:\[|$))')

# Parse columm headers, handling possible defaults
_COL_HEADER_RE = re.compile(r"^\s*?(?P<name>.*?)"
                            r"(\s*?=\s*?'(?P<default>.*?)'\s*?){0,1}$")


def _parse_col_descriptor(column_descriptor):
    components = [c.strip() for c in column_descriptor.split(",")]

    columns = []
    defaults = []

    for column in components:
        m = _COL_HEADER_RE.search(column)

        if m is None:
            raise ValueError("'%s' is not a valid column header" % column)

        name, default = m.group('name', 'default')

        columns.append(name)
        defaults.append(default)

    return columns, defaults


def _parse_header(header):
    format_str, col_desc = (c.strip() for c in header.split("=", 1))

    if format_str != "Format":
        raise ValueError("'%s' does not appear to be a wsclean header")

    return _parse_col_descriptor(col_desc)


def _parse_lines(fh, line_nr, column_names, defaults, converters):
    source_data = [[] for _ in range(len(column_names))]

    for line_nr, line in enumerate(fh, line_nr):
        components = [c.strip() for c in re.split(_COMMA_SPLIT_RE, line)]

        if len(components) != len(column_names):
            raise ValueError("line %d '%s' should have %d components" %
                             (line_nr, line, len(column_names)))

        # Iterate through each column's data
        it = zip(column_names, components, converters, source_data, defaults)

        for name, comp, conv, data_list, default in it:
            if not comp:
                if default is None:
                    try:
                        default = conv()
                    except Exception as e:
                        raise ValueError("No value supplied for column '%s' "
                                         "on line %d and no default was "
                                         "supplied either. Attempting to "
                                         "generate a default produced the "
                                         "following exception %s"
                                         % (name, line_nr, str(e)))

                value = default
            else:
                value = comp

            data_list.append(conv(value))

    return zip(*(column_names, source_data))


def load(filename):
    """
    Loads wsclean component model.

    .. code-block:: python

        sources = load("components.txt")
        sources = dict(sources)  # Convert to dictionary

        I = sources["I"]
        ref_freq = sources["ReferenceFrequency"]

    See the `WSClean Component List
    <https://sourceforge.net/p/wsclean/wiki/ComponentList/>`_
    for further details.

    Parameters
    ----------
    filename : str or iterable
        Filename of wsclean model file or iterable
        producing the lines of the file.

    See Also
    --------
    africanus.model.wsclean.spectra

    Returns
    -------
    list of (name, list of values) tuples
        list of column (name, value) tuples
    """

    if isinstance(filename, str):
        fh = open(filename, "r")
        fh = iter(fh)
        close_filename = True
    else:
        fh = iter(filename)
        close_filename = False

    try:
        # Search for a header until we find a non-empty string
        header = ''
        line_nr = 1

        for headers in fh:
            header = headers.split("#", 1)[0].strip()

            if header:
                break

            line_nr += 1

        if not header:
            raise ValueError("'%s' does not contain a valid wsclean header"
                             % filename)

        column_names, defaults = _parse_header(header)

        try:
            converters = [_COLUMN_CONVERTERS[n] for n in column_names]
        except KeyError as e:
            raise ValueError("No converter registered for column %s" % str(e))

        return _parse_lines(fh, line_nr, column_names, defaults, converters)
    finally:
        if close_filename:
            fh.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `codex-africanus` package."""

import numpy as np

import pytest


def rf(*a, **kw):
    return np.random.random(*a, **kw)


def rc(*a, **kw):
    return rf(*a, **kw) + 1j*rf(*a, **kw)


def test_phase_delay():
    from africanus.rime import phase_delay

    uvw = np.random.random(size=(100, 3))
    lm = np.random.random(size=(10, 2))
    frequency = np.linspace(.856e9, .856e9*2, 64, endpoint=True)

    from africanus.constants import minus_two_pi_over_c

    # Test complex phase at a particular index in the output
    uvw_i, lm_i, freq_i = 2, 3, 5

    u, v, w = [1, 2, 3]
    l, m = [0.1, 0.2]
    freq = 0.856e9

    # Set up values in the input
    uvw[uvw_i] = [u, v, w]
    lm[lm_i] = [l, m]
    frequency[freq_i] = freq

    # Compute complex phase
    complex_phase = phase_delay(lm, uvw, frequency)

    # Test singular value vs a point in the output
    n = np.sqrt(1.0 - l**2 - m**2) - 1.0
    phase = minus_two_pi_over_c*(u*l + v*m + w*n)*freq
    assert np.all(np.exp(1j*phase) == complex_phase[lm_i, uvw_i, freq_i])


def test_feed_rotation():
    import numpy as np
    from africanus.rime import feed_rotation

    parangles = np.random.random((10, 5))
    pa_sin = np.sin(parangles)
    pa_cos = np.cos(parangles)

    fr = feed_rotation(parangles, feed_type='linear')
    np_expr = np.stack([pa_cos, pa_sin, -pa_sin, pa_cos], axis=2)
    assert np.allclose(fr, np_expr.reshape(10, 5, 2, 2))

    fr = feed_rotation(parangles, feed_type='circular')
    zeros = np.zeros_like(pa_sin)
    np_expr = np.stack([pa_cos - 1j*pa_sin, zeros,
                        zeros, pa_cos + 1j*pa_sin], axis=2)
    assert np.allclose(fr, np_expr.reshape(10, 5, 2, 2))


def test_dask_phase_delay():
    da = pytest.importorskip('dask.array')
    from africanus.rime import phase_delay as np_phase_delay
    from africanus.rime.dask import phase_delay as dask_phase_delay

    # So that 1 > 1 - l**2 - m**2 >= 0
    lm = np.random.random(size=(10, 2))*0.01
    uvw = np.random.random(size=(100, 3))
    frequency = np.linspace(.856e9, .856e9*2, 64, endpoint=True)

    dask_lm = da.from_array(lm, chunks=(5, 2))
    dask_uvw = da.from_array(uvw, chunks=(25, 3))
    dask_frequency = da.from_array(frequency, chunks=16)

    dask_phase = dask_phase_delay(dask_lm, dask_uvw, dask_frequency)
    np_phase = np_phase_delay(lm, uvw, frequency)

    # Should agree completely
    assert np.all(np_phase == dask_phase.compute())


def test_dask_feed_rotation():
    da = pytest.importorskip('dask.array')
    import numpy as np
    from africanus.rime import feed_rotation as np_feed_rotation
    from africanus.rime.dask import feed_rotation

    parangles = np.random.random((10, 5))
    dask_parangles = da.from_array(parangles, chunks=(5, (2, 3)))

    np_fr = np_feed_rotation(parangles, feed_type='linear')
    assert np.all(np_fr == feed_rotation(dask_parangles, feed_type='linear'))

    np_fr = np_feed_rotation(parangles, feed_type='circular')
    assert np.all(np_fr == feed_rotation(dask_parangles, feed_type='circular'))



import numpy as np


from africanus.util.numba import jit


@jit(nogil=True, nopython=True, cache=True)
def fac(x):
    if x < 0:
        raise ValueError("Factorial input is negative.")
    if x == 0:
        return 1
    factorial = 1
    for i in range(1, x + 1):
        factorial *= i
    return factorial


@jit(nogil=True, nopython=True, cache=True)
def pre_fac(k, n, m):
    numerator = (-1.0)**k * fac(n-k)
    denominator = (fac(k) * fac((n+m)/2.0 - k) * fac((n-m)/2.0 - k))
    return numerator / denominator


@jit(nogil=True, nopython=True, cache=True)
def zernike_rad(m, n, rho):
    if (n < 0 or m < 0 or abs(m) > n):
        raise ValueError("m and n values are incorrect.")
    radial_component = 0
    for k in range((n-m)/2+1):
        radial_component += pre_fac(k, n, m) * rho ** (n - 2.0 * k)
    return radial_component


@jit(nogil=True, nopython=True, cache=True)
def zernike(j, rho, phi):
    if rho > 1:
        return 0 + 0j
    j += 1
    n = 0
    j1 = j-1
    while (j1 > n):
        n += 1
        j1 -= n
    m = (-1)**j * ((n % 2) + 2 * int((j1+((n+1) % 2)) / 2.0))
    if (m > 0):
        return zernike_rad(m, n, rho) * np.cos(m * phi)
    if (m < 0):
        return zernike_rad(-m, n, rho) * np.sin(-m * phi)
    return zernike_rad(0, n, rho)


@jit(nogil=True, nopython=True, cache=True)
def _convert_coords(l, m):
    rho, phi = (l**2 + m ** 2) ** 0.5, np.arctan2(l, m)
    return rho, phi


@jit(nogil=True, nopython=True, cache=True)
def nb_zernike_dde(coords, coeffs, noll_index, out):
    sources, times, ants, chans, corrs = out.shape
    npoly = coeffs.shape[-1]

    for s in range(sources):
        for t in range(times):
            for a in range(ants):
                for c in range(chans):
                    l, m, freq = coords[:, s, t, a, c]
                    rho, phi = _convert_coords(l, m)

                    for co in range(corrs):
                        zernike_sum = 0

                        for p in range(npoly):
                            zc = coeffs[a, c, co, p]
                            zn = noll_index[a, c, co, p]
                            zernike_sum += zc * zernike(zn, rho, phi)

                        out[s, t, a, c, co] = zernike_sum

    return out


def zernike_dde(coords, coeffs, noll_index):
    """ Wrapper for :func:`nb_zernike_dde` """
    _, sources, times, ants, chans = coords.shape
    # ant, chan, corr_1, ..., corr_n, poly
    corr_shape = coeffs.shape[2:-1]
    npoly = coeffs.shape[-1]

    # Flatten correlation dimensions for numba function
    fcorrs = np.product(corr_shape)
    ddes = np.empty((sources, times, ants, chans, fcorrs), coeffs.dtype)

    coeffs = coeffs.reshape((ants, chans, fcorrs, npoly))
    noll_index = noll_index.reshape((ants, chans, fcorrs, npoly))

    result = nb_zernike_dde(coords, coeffs, noll_index, ddes)

    # Reshape to full correlation size
    return result.reshape((sources, times, ants, chans) + corr_shape)


_ZERNICKE_DOCSTRING = (
    """
Computes Direction Dependent Effects by evaluating
`Zernicke Polynomials <zernike_wiki_>`_
defined by coefficients ``coeffs``
and noll indexes ``noll_index``
at the specified coordinates ``coords``.

Decomposition of a voxel beam cube into Zernicke
polynomial coefficients can be achieved through the
use of the eidos_ package.

.. _zernike_wiki: https://en.wikipedia.org/wiki/Zernike_polynomials
.. _eidos: https://github.com/kmbasad/eidos/

Parameters
---------------
coords : :class:`numpy.ndarray`
   Float coordinates at which to evaluate the zernike polynomials.
   Has shape :code:`(3, source, time, ant, chan)`. The three components in
   the first dimension represent
   l, m and frequency coordinates, respectively.
coeffs : :class:`numpy.ndarray`
  complex Zernicke polynomial coefficients.
  Has shape :code:`(ant, chan, corr_1, ..., corr_n, poly)`
  where ``poly`` is the number of polynomial coefficients
  and ``corr_1, ..., corr_n`` are a variable number of
  correlation dimensions.
noll_index : :class:`numpy.ndarray`
  Noll index associated with each polynomial coefficient.
  Has shape :code:`(ant, chan, corr_1, ..., corr_n, poly)`.

Returns
----------
dde : :class:`numpy.ndarray`
   complex values with shape
   :code:`(source, time, ant, chan, corr_1, ..., corr_n)`
""")

zernike_dde.__doc__ = _ZERNICKE_DOCSTRING



# flake8: noqa

from africanus.coordinates.coordinates import radec_to_lmn
from africanus.coordinates.coordinates import radec_to_lm
from africanus.coordinates.coordinates import lmn_to_radec
from africanus.coordinates.coordinates import lm_to_radec



# -*- coding: utf-8 -*-


from hashlib import sha1


def sha_hash_file(filename):
    """ Compute the SHA1 hash of filename """
    hash_sha = sha1()

    with open(filename, 'rb') as f:
        for chunk in iter(lambda: f.read(1024*1024), b""):
            hash_sha.update(chunk)

    return hash_sha.hexdigest()



# -*- coding: utf-8 -*-


from pathlib import Path

import numpy as np

from africanus.util.beams import beam_filenames
from africanus.util.requirements import requires_optional

try:
    from astropy.io import fits
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None


DEFAULT_SCHEMA = Path("test_beam_$(corr)_$(reim).fits")
LINEAR_CORRELATIONS = [9, 10, 11, 12]
CIRCULAR_CORRELATIONS = [5, 6, 7, 8]


BITPIX_MAP = {
    np.int8: 8,
    np.uint8: 8,
    np.int16: 16,
    np.uint16: 16,
    np.int32: 32,
    np.uint32: 32,
    np.float32: -32,
    np.float64: -64,
}


@requires_optional('astropy.io.fits', opt_import_error)
def beam_factory(polarisation_type='linear',
                 frequency=None,
                 npix=257,
                 dtype=np.float64,
                 schema=DEFAULT_SCHEMA,
                 overwrite=True):
    """ Generate a MeqTrees compliant beam cube """

    if npix % 2 != 1:
        raise ValueError("npix '%d' must be odd" % npix)

    # MeerKAT l-band, 64 channels
    if frequency is None:
        frequency = np.linspace(.856e9, .856e9*2, 64,
                                endpoint=True, dtype=np.float64)

    # Generate a linear space of grid frequencies
    gfrequency = np.linspace(frequency[0], frequency[-1],
                             33, dtype=np.float64)
    bandwidth = gfrequency[-1] - frequency[0]
    bandwidth_delta = bandwidth / gfrequency.shape[0]-1

    if polarisation_type == 'linear':
        CORR = LINEAR_CORRELATIONS
    elif polarisation_type == 'circular':
        CORR = CIRCULAR_CORRELATIONS
    else:
        raise ValueError("Invalid polarisation_type %s" % polarisation_type)

    extent_deg = 3.0
    coords = np.linspace(-extent_deg/2, extent_deg/2, npix, endpoint=True)

    crpix = 1 + npix // 2            # Reference pixel (FORTRAN)
    crval = coords[crpix - 1]        # Reference value
    cdelt = extent_deg / (npix - 1)  # Delta

    # List of key values of the form:
    #
    #    (key, None)
    #    (key, (value,))
    #    (key, (value, comment))
    #
    # We put them in a list so that they are added to the
    # FITS header in the correct order
    axis1 = [
        ("CTYPE", ('X', "points right on the sky")),
        ("CUNIT", ('DEG', 'degrees')),
        ("NAXIS", (npix, "number of X")),
        ("CRPIX", (crpix, "reference pixel (one relative)")),
        ("CRVAL", (crval, "degrees")),
        ("CDELT", (cdelt, "degrees"))]

    axis2 = [
        ("CTYPE", ('Y', "points up on the sky")),
        ("CUNIT", ('DEG', 'degrees')),
        ("NAXIS", (npix, "number of Y")),
        ("CRPIX", (crpix, "reference pixel (one relative)")),
        ("CRVAL", (crval, "degrees")),
        ("CDELT", (cdelt, "degrees"))]

    axis3 = [
        ("CTYPE", ('FREQ', )),
        ("CUNIT", None),
        ("NAXIS", (gfrequency.shape[0], "number of FREQ")),
        ("CRPIX", (1, "reference frequency position")),
        ("CRVAL", (gfrequency[0], "reference frequency")),
        ("CDELT", (bandwidth_delta, "frequency step in Hz"))]

    axes = [axis1, axis2, axis3]

    metadata = [
        ('SIMPLE', True),
        ('BITPIX', BITPIX_MAP[dtype]),
        ('NAXIS', len(axes)),
        ('OBSERVER', "Astronomer McAstronomerFace"),
        ('ORIGIN', "Artificial"),
        ('TELESCOP', "Telescope"),
        ('OBJECT', 'beam'),
        ('EQUINOX', 2000.0),
    ]

    # Create header and set metadata
    header = fits.Header(metadata)

    # Now set the key value entries for each axis
    ax_info = [('%s%d' % (k, a),) + vt
               for a, axis_data in enumerate(axes, 1)
               for k, vt in axis_data
               if vt is not None]
    header.update(ax_info)

    # Now setup the GFREQS
    # Jitter them randomly, except for the endpoints
    frequency_jitter = np.random.random(size=gfrequency.shape)-0.5
    frequency_jitter *= 0.1*bandwidth_delta
    frequency_jitter[0] = frequency_jitter[-1] = 0.0
    gfrequency += frequency_jitter

    # Check that gfrequency is monotically increasing
    assert np.all(np.diff(gfrequency) >= 0.0)

    for i, gfreq in enumerate(gfrequency, 1):
        header['GFREQ%d' % i] = gfreq

    # Figure out the beam filenames from the schema
    filenames = beam_filenames(str(schema), CORR)

    # Westerbork beam model
    coords = np.deg2rad(coords)
    r = np.sqrt(coords[None, :, None]**2 + coords[None, None, :]**2)
    fq = gfrequency[:, None, None]
    beam = np.cos(np.minimum(65*fq*1e-9*r, 1.0881))**3

    for filename in [f for ri_pair in filenames.values() for f in ri_pair]:
        primary_hdu = fits.PrimaryHDU(beam, header=header)
        primary_hdu.writeto(filename, overwrite=overwrite)

    return filenames


if __name__ == "__main__":
    beam_factory()




# -*- coding: utf-8 -*-


try:
    from dask.array import Array as dask_array
except ImportError:
    dask_array = object()

import numba
import numpy as np


def _numpy_dtype(arg):
    if isinstance(arg, np.ndarray):
        return arg.dtype
    elif isinstance(arg, numba.types.npytypes.Array):
        return np.dtype(arg.dtype.name)
    elif isinstance(arg, dask_array):
        return arg.dtype
    else:
        raise ValueError("Unhandled type %s" % type(arg))


def infer_complex_dtype(*args):
    """ Infer complex datatype from arg inputs """
    return np.result_type(np.complex64, *(_numpy_dtype(a) for a in args))



# -*- coding: utf-8 -*-


from os.path import join as pjoin

from appdirs import AppDirs

from africanus import __version__

_dirs = AppDirs("codex-africanus", "radio-astronomer", __version__)

user_data_dir = _dirs.user_data_dir
downloads_dir = pjoin(user_data_dir, "downloads")
include_dir = pjoin(user_data_dir, "include")

del __version__
del _dirs



# -*- coding: utf-8 -*-


from decorator import decorate
from numba import types

from africanus.util.docs import on_rtd

if on_rtd():
    # Fake decorators when on readthedocs
    def _fake_decorator(*args, **kwargs):
        def decorator(fn):
            def wrapper(*args, **kwargs):
                return fn(*args, **kwargs)

            return decorate(fn, wrapper)

        return decorator

    cfunc = _fake_decorator
    jit = _fake_decorator
    generated_jit = _fake_decorator
    njit = _fake_decorator
    stencil = _fake_decorator

else:
    from numba import cfunc, jit, njit, generated_jit, stencil  # noqa


def is_numba_type_none(arg):
    """
    Returns True if the numba type represents None


    Parameters
    ----------
    arg : :class:`numba.Type`
        The numba type

    Returns
    -------
    boolean
        True if the type represents None
    """
    return (isinstance(arg, types.misc.NoneType) or
            (isinstance(arg, types.misc.Omitted) and arg.value is None))



# -*- coding: utf-8 -*-

"""Top-level package for Codex Africanus."""


# NOTE(sjperkins)
# Imports at this level within this module should be avoided,
# or should fail gracefully as this is the base africanus module.
# The setup.py file accesses the ``africanus.install`` modules
import africanus.util.jax_init  # noqa

__author__ = """Simon Perkins"""
__email__ = 'sperkins@ska.ac.za'
__version__ = '0.2.0'



# -*- coding: utf-8 -*-


import numpy as np


def estimate_kaiser_bessel_beta(W):
    r"""
    Estimate the kaiser bessel beta using the following heuristic:

    .. math::

        \beta = 2.34 \times W

    Derived from `Nonuniform fast Fourier transforms
    using min-max interpolation
    <nufft-min-max-ref_>`_.

    .. _nufft-min-max-ref: https://ieeexplore.ieee.org/document/1166689/

    Parameters
    ----------
    W : int
        Width of the filter

    Returns
    -------
    float
        Kaiser Bessel beta shape parameter
    """
    return 2.34*W


def kaiser_bessel(u, W, beta):
    r"""
    Compute a 1D Kaiser Bessel filter as defined
    in `Selection of a Convolution Function
    for Fourier Inversion Using Gridding
    <kbref_>`_.

    .. _kbref: https://ieeexplore.ieee.org/document/97598/

    Parameters
    ----------
    u : :class:`numpy.ndarray`
        Filter positions
    W : int
        Width of the filter
    beta : float, optional
        Kaiser Bessel shape parameter

    Returns
    -------
    :class:`numpy.ndarray`
        Kaiser Bessel filter with the same shape as `u`
    """

    # Sanity check
    hW = W // 2
    assert np.all(-hW <= u) & np.all(u <= hW)

    param = 1 - (2 * u / W)**2
    param[param < 0] = 0  # Zero negative numbers
    return np.i0(beta * np.sqrt(param)) / np.i0(beta)


def kaiser_bessel_with_sinc(u, W, oversample, beta, normalise=True):
    """
    Produces a filter composed of Kaiser Bessel multiplied by a sinc.

    Accounts for the oversampling factor, as well as normalising the filter.

    Parameters
    ----------
    u : :class:`numpy.ndarray`
        Filter positions
    W : int
        Width of the filter
    oversample : int
        Oversampling factor
    beta : float
        Kaiser Bessel shape parameter
    normalise : optional, {True, False}
        True if the filter should be normalised

    Returns
    -------
    :class:`numpy.ndarray`
        Filter with the same shape as `u`
    """
    kb = kaiser_bessel(u, W, beta)
    kb *= oversample
    kb *= np.sinc(u / oversample)

    if normalise:
        kb /= np.trapz(kb, u)

    return kb


def kaiser_bessel_fourier(x, W, beta):
    r"""
    Computes the Fourier Transform of a 1D Kaiser Bessel filter.
    as defined in `Selection of a Convolution Function
    for Fourier Inversion Using Gridding
    <kbref_>`_.

    .. _kbref: https://ieeexplore.ieee.org/document/97598/


    Parameters
    ----------
    x : :class:`numpy.ndarray`
        Filter positions
    W : int
        Width of the filter.
    beta : float
        Kaiser bessel shape parameter

    Returns
    -------
    :class:`numpy.ndarray`
        Fourier Transform of the Kaiser Bessel,
        with the same shape as `x`.
    """
    term = (np.pi*W*x)**2 - beta**2
    val = np.lib.scimath.sqrt(term).real
    return np.sin(val)/val



# -*- coding: utf-8 -*-


from functools import reduce
import logging
from operator import mul
from os.path import join as pjoin

import numpy as np

from africanus.model.coherency.conversion import (_element_indices_and_shape,
                                                  CONVERT_DOCS,
                                                  MissingConversionInputs)
from africanus.util.code import memoize_on_key, format_code
from africanus.util.cuda import cuda_type, grids
from africanus.util.jinja2 import jinja_env
from africanus.util.requirements import requires_optional

try:
    import cupy as cp
    from cupy.cuda.compiler import CompileException
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None

log = logging.getLogger(__name__)

stokes_conv = {
    'RR': {('I', 'V'): ("complex", "make_{{out_type}}({{I}} + {{V}}, 0)")},
    'RL': {('Q', 'U'): ("complex", "make_{{out_type}}({{Q}}, {{U}})")},
    'LR': {('Q', 'U'): ("complex", "make_{{out_type}}({{Q}}, -{{U}})")},
    'LL': {('I', 'V'): ("complex", "make_{{out_type}}({{I}} - {{V}}, 0)")},

    'XX': {('I', 'Q'): ("complex", "make_{{out_type}}({{I}} + {{Q}}, 0)")},
    'XY': {('U', 'V'): ("complex", "make_{{out_type}}({{U}}, {{V}})")},
    'YX': {('U', 'V'): ("complex", "make_{{out_type}}({{U}}, -{{V}})")},
    'YY': {('I', 'Q'): ("complex", "make_{{out_type}}({{I}} - {{Q}}, 0)")},

    'I': {('XX', 'YY'): ("real", "(({{XX}}.x + {{YY}}.x) / 2)"),
          ('RR', 'LL'): ("real", "(({{RR}}.x + {{LL}}.x) / 2)")},

    'Q': {('XX', 'YY'): ("real", "(({{XX}}.x - {{YY}}.x) / 2)"),
          ('RL', 'LR'): ("real", "(({{RL}}.x + {{LR}}.x) / 2)")},

    'U': {('XY', 'YX'): ("real", "(({{XY}}.x + {{YX}}.x) / 2)"),
          ('RL', 'LR'): ("real", "(({{RL}}.y - {{LR}}.y) / 2)")},

    'V': {('XY', 'YX'): ("real", "(({{XY}}.y - {{YX}}.y) / 2)"),
          ('RR', 'LL'): ("real", "(({{RR}}.x - {{LL}}.x) / 2)")},
}


def stokes_convert_setup(input, input_schema, output_schema):
    input_indices, input_shape = _element_indices_and_shape(input_schema)
    output_indices, output_shape = _element_indices_and_shape(output_schema)

    if input.shape[-len(input_shape):] != input_shape:
        raise ValueError("Last dimension of input doesn't match input schema")

    mapping = []
    dtypes = []

    # Figure out how to produce an output from available inputs
    for okey, out_idx in output_indices.items():
        try:
            deps = stokes_conv[okey]
        except KeyError:
            raise ValueError("Unknown output '%s'. Known types '%s'"
                             % (okey, stokes_conv.keys()))

        found_conv = False

        # Find a mapping for which we have inputs
        for (c1, c2), (dtype, fn) in deps.items():
            # Get indices for both correlations
            try:
                c1_idx = input_indices[c1]
            except KeyError:
                continue

            try:
                c2_idx = input_indices[c2]
            except KeyError:
                continue

            found_conv = True
            dtypes.append(dtype)

            mapping.append(((c1, c1_idx), (c2, c2_idx), out_idx, fn))
            break

        # We must find a conversion
        if not found_conv:
            raise MissingConversionInputs("None of the supplied inputs '%s' "
                                          "can produce output '%s'. It can be "
                                          "produced by the following "
                                          "combinations '%s'." % (
                                                input_schema,
                                                okey, deps.keys()))

    # Output types must be all "real" or all "complex"
    if not all(dtypes[0] == dt for dt in dtypes[1:]):
        raise ValueError("Output data types differ %s" % dtypes)

    return mapping, input_shape, output_shape, dtypes[0]


def schema_to_tuple(schema):
    if isinstance(schema, (tuple, list)):
        return tuple(schema_to_tuple(s) for s in schema)
    else:
        return schema


def _key_fn(inputs, input_schema, output_schema):
    return (inputs.dtype,
            schema_to_tuple(input_schema),
            schema_to_tuple(output_schema))


_TEMPLATE_PATH = pjoin("model", "coherency", "cuda", "conversion.cu.j2")


@memoize_on_key(_key_fn)
def _generate_kernel(inputs, input_schema, output_schema):
    mapping, in_shape, out_shape, out_dtype = stokes_convert_setup(
                                                inputs,
                                                input_schema,
                                                output_schema)

    # Flatten input and output shapes
    # Check that number elements are the same
    in_elems = reduce(mul, in_shape, 1)
    out_elems = reduce(mul, out_shape, 1)

    if in_elems != out_elems:
        raise ValueError("Number of input_schema elements %s "
                         "and output schema elements %s "
                         "must match for CUDA kernel." %
                         (in_shape, out_shape))

    # Infer the output data type
    if out_dtype == "real":
        if np.iscomplexobj(inputs):
            out_dtype = inputs.real.dtype
        else:
            out_dtype = inputs.dtype
    elif out_dtype == "complex":
        if np.iscomplexobj(inputs):
            out_dtype = inputs.dtype
        else:
            out_dtype = np.result_type(inputs.dtype, np.complex64)
    else:
        raise ValueError("Invalid setup dtype %s" % out_dtype)

    cuda_out_dtype = cuda_type(out_dtype)
    assign_exprs = []

    # Render the assignment expression for each element
    for (c1, c1i), (c2, c2i), outi, template_fn in mapping:
        # Flattened indices
        flat_outi = np.ravel_multi_index(outi, out_shape)
        render = jinja_env.from_string(template_fn).render
        kwargs = {c1: "in[%d]" % np.ravel_multi_index(c1i, in_shape),
                  c2: "in[%d]" % np.ravel_multi_index(c2i, in_shape),
                  "out_type": cuda_out_dtype}

        expr_str = render(**kwargs)
        assign_exprs.append("out[%d] = %s;" % (flat_outi, expr_str))

    # Now render the main template
    render = jinja_env.get_template(_TEMPLATE_PATH).render
    name = "stokes_convert"
    code = render(kernel_name=name,
                  input_type=cuda_type(inputs.dtype),
                  output_type=cuda_type(out_dtype),
                  assign_exprs=assign_exprs,
                  elements=in_elems).encode("utf-8")

    # cuda block, flatten non-schema dims into a single source dim
    blockdimx = 512
    block = (blockdimx, 1, 1)

    return (cp.RawKernel(code, name), block, in_shape, out_shape, out_dtype)


@requires_optional('cupy', opt_import_error)
def convert(inputs, input_schema, output_schema):
    (kernel, block,
     in_shape, out_shape, dtype) = _generate_kernel(inputs,
                                                    input_schema,
                                                    output_schema)

    # Flatten non-schema input dimensions,
    # from inspection of the cupy reshape code,
    # this incurs a copy when inputs is non-contiguous
    nsrc = reduce(mul, inputs.shape[:-len(in_shape)], 1)
    nelems = reduce(mul, in_shape, 1)

    rinputs = inputs.reshape(nsrc, nelems)
    assert rinputs.flags.c_contiguous
    grid = grids((nsrc, 1, 1), block)

    outputs = cp.empty(shape=rinputs.shape, dtype=dtype)

    try:
        kernel(grid, block, (rinputs, outputs))
    except CompileException:
        log.exception(format_code(kernel.code))
        raise

    shape = inputs.shape[:-len(in_shape)] + out_shape
    outputs = outputs.reshape(shape)
    assert outputs.flags.c_contiguous
    return outputs


try:
    convert.__doc__ = CONVERT_DOCS.substitute(
                                array_type=":class:`cupy.ndarray`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


import numpy as np
from numpy.testing import assert_array_almost_equal
import pytest

from africanus.model.spectral.spec_model import (spectral_model,
                                                 numpy_spectral_model)


@pytest.fixture
def flux():
    def impl(nsrc):
        return np.random.random(nsrc)

    return impl


@pytest.fixture
def frequency():
    def impl(nchan):
        return np.linspace(.856e9, 2*.856e9, nchan)

    return impl


@pytest.fixture
def ref_freq():
    def impl(shape):
        return np.full(shape, 3*.856e9/2)

    return impl


@pytest.fixture
def spi():
    def impl(shape):
        return 0.7 + np.random.random(shape) * 0.2

    return impl


@pytest.mark.parametrize("base", [0, 1, 2, "std", "log", "log10",
                                  ["log", "std", "std", "std"]])
@pytest.mark.parametrize("npol", [0, 1, 2, 4])
def test_spectral_model_multiple_spi(flux, ref_freq, frequency, base, npol):
    nsrc = 10
    nchan = 16
    nspi = 6

    # Clip to number of polarisations
    if isinstance(base, list):
        base = base[0] if npol == 0 else base[:npol]

    if npol > 0:
        flux_slice = (slice(None), None)
        flux_shape = (nsrc, npol)
        spi_shape = (nsrc, nspi, npol)
    else:
        flux_slice = (slice(None),)
        flux_shape = (nsrc,)
        spi_shape = (nsrc, nspi)

    I = flux(nsrc)  # noqa
    stokes = np.broadcast_to(I[flux_slice], flux_shape)  # noqa
    spi = 0.7 + np.random.random(spi_shape) * 0.2
    ref_freq = ref_freq(nsrc)
    freq = frequency(nchan)

    model = spectral_model(stokes, spi, ref_freq, freq, base=base)
    np_model = numpy_spectral_model(stokes, spi, ref_freq, freq, base=base)

    assert_array_almost_equal(model, np_model)
    assert model.flags.c_contiguous is True


@pytest.mark.parametrize("base", [0, 1, 2, "std", "log", "log10",
                                  ["log", "std", "std", "std"]])
@pytest.mark.parametrize("npol", [0, 1, 2, 4])
def test_dask_spectral_model(flux, ref_freq, frequency, base, npol):
    da = pytest.importorskip("dask.array")
    from africanus.model.spectral.spec_model import (
                        spectral_model as np_spectral_model)
    from africanus.model.spectral.dask import spectral_model

    sc = (5, 5)
    fc = (8, 8)
    spc = (6,)

    nsrc = sum(sc)
    nchan = sum(fc)
    nspi = sum(spc)

    # Clip to number of polarisations
    if isinstance(base, list):
        base = base[0] if npol == 0 else base[:npol]

    if npol > 0:
        flux_slice = (slice(None), None)
        flux_shape = (nsrc, npol)
        spi_shape = (nsrc, nspi, npol)
        cc = (npol,)
    else:
        flux_slice = (slice(None),)
        flux_shape = (nsrc,)
        spi_shape = (nsrc, nspi)
        cc = ()

    I = flux(nsrc)  # noqa
    stokes = np.broadcast_to(I[flux_slice], flux_shape)  # noqa
    spi = 0.7 + np.random.random(spi_shape) * 0.2
    ref_freq = ref_freq(nsrc)
    freq = frequency(nchan)

    da_stokes = da.from_array(stokes, chunks=(sc,) + cc)
    da_spi = da.from_array(spi, chunks=(sc, spc) + cc)
    da_ref_freq = da.from_array(ref_freq, chunks=sc)
    da_freq = da.from_array(freq, chunks=fc)

    da_model = spectral_model(da_stokes, da_spi,
                              da_ref_freq, da_freq, base=base)

    np_model = np_spectral_model(stokes, spi, ref_freq, freq, base=base)
    assert_array_almost_equal(da_model, np_model)



# -*- coding: utf-8 -*-


from africanus.model.shape.gaussian_shape import (gaussian as np_gaussian,
                                                  GAUSSIAN_DOCS)
from africanus.util.requirements import requires_optional

import numpy as np

try:
    import dask.array as da
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None


def _wrapper(uvw, frequency, shape_params):
    return np_gaussian(uvw[0], frequency, shape_params[0])


@requires_optional('dask.array', opt_import_error)
def gaussian(uvw, frequency, shape_params):
    dtype = np.result_type(uvw.dtype, frequency.dtype, shape_params.dtype)

    return da.blockwise(_wrapper, ("source", "row", "chan"),
                        uvw, ("row", "uvw-comp"),
                        frequency, ("chan",),
                        shape_params, ("source", "shape-comp"),
                        dtype=dtype)


try:
    gaussian.__doc__ = GAUSSIAN_DOCS.substitute(
                            array_type=":class:`dask.array.Array`")
except AttributeError:
    pass
